/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Identify candidate lightcurves. See Section 5 of the report.
 Note that homebrew and MACHO are not discussed in the report; the latter of these is currently incomplete.
 
 Started: 19/03/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include <stdio.h>
#include <math.h>
#include "config.h"
#include "fit_base.c"

int identify_candidate(double data[3][N], double median, int num_data_points)
{
	double sum, mean, F_mean, F_sigma;
	double delta;
	int event_1_start, event_1_end;
	int event_2_start, event_2_end;
	int event_type;
	int window_length, window_start;
	int num_segments, num_outside;
	double variance_array[N], variance_sum, temp;
	int i, j, k, l, m, a, status;
	int min;
	float I, J;
	double lsfit_params[5], lsfit_errors[5];
	double chi_global2, chi2;
	int num_points_local;
	double peak_data[3][N];
	double A;
	double base_fit_params[NUM_BASE_PARAMETERS], base_fit_errors[NUM_BASE_PARAMETERS];
	double base_chi2, chi_max;
	
	if (USE_OGLE_ID)
	{
		//
		// Use the method for detecting candidates as used by OGLEII, described in Wozniak et. al, 2001.
		//
		
		/*
		 Determine whether the lightcurve would be classified as "Transient" or "Continuous" - OGLEII only considers the former. Using Wozniak (2000) section 3.9
		 Count the number of points departing at least 3 sigma upwards, and the same downwards, and work out their ratio (n_up / n_down) - this should be outside the region 0.5 - 2.0 for a transient event.
		 */
		// j keeps track of the number of points below, k the number of points above.
		for (i = j = k = 0; i < num_data_points; i++)
		{
			if (data[1][i] > (median + 3 * data[2][i]))
			{
				// We've got a decrease in the flux (an increase in the magnitude)
				j++;
			}
			else if (data[1][i] < (median - 3 * data[2][i]))
			{
				// We've got an increase in the flux (a decrease in the magnitude)
				k++;
			};
		};
		if ((k / j) < 2.0 && (k / j) > 0.5)
		{
			// We have a continuous variable - dismiss it.
			if (DEBUG_CANDIDATES)
			{
				printf("Continuous variable");
			}
#if MONTE_CARLO
			mc_fail_array[0]++;
#endif
			return 0;
		};
		
		/*
		 Find a running window of 50% of the data points (consecutively) which minimises sigma_f_base around the mean f_base - looping the data so that we don't discriminate against the end points.
		 */
		
		// First, get half the number of data points
		window_length = num_data_points * OGLE_WINDOW_LENGTH;

		// Go through each of the data points as the starting point
		for (i = 0; i < num_data_points; i++)
		{
			// Calculate the mean for this window
			sum = 0;
			for (j = 0; j < window_length; j++)
			{
				if (i + j < num_data_points)
				{
					sum += data[1][i + j];
				}
				else
				{
					sum += data[1][i + j - num_data_points];
				}
			};
			F_mean = sum / window_length;
			
			// We want to start at the ith data point, and go through the data set (looping round) working out the variance.
			sum = 0;
			for (j = 0; j < window_length; j++)
			{
				if (i + j < num_data_points)
				{
					sum += pow(data[1][i + j] - F_mean, 2.0);
				}
				else
				{
					sum += pow(data[1][i + j - num_data_points] - F_mean, 2.0);
				}
			};
			variance_array[i] = sum / (window_length - 1);
		}
		// We want the one with the minimum variance.
		window_start = gsl_stats_min_index(variance_array, 1, num_data_points);
		F_sigma = sqrt(variance_array[window_start]);
		
		/*
		 Split running window up into 10 point blocks, calculate the mean F and sigma_F in each, and require that all but a maximum of 2 of the means are within sigma_f_base (F_sigma) of f_base (F_mean). Also require the mean of the sigma_F to be greater than 0.5 * sigma_f_base.
		 */
		num_segments = (window_length - (window_length % 10)) / 10;
		num_outside = 0;
		for (j = 0; j < num_segments; j++)
		{
			// Calculate the mean
			sum = 0;
			for (k = 0; k < 10; k++)
			{
				if (window_start + k + 10 * j < num_data_points)
				{
					sum += data[1][window_start + k + 10 * j];
				}
				else
				{
					sum += data[1][window_start + k + 10 * j - num_data_points];
				}
			};
			mean = sum / 10.0;
			if (mean > F_mean + F_sigma || mean < F_mean - F_sigma)
			{
				num_outside++;
				// If we've got to more than 2 of num_outside, return false
				if (num_outside > 2)
				{
					if (DEBUG_CANDIDATES)
					{
						printf("More than 2 segments outside range");
					}
#if MONTE_CARLO
					mc_fail_array[1]++;
#endif
					return 0;
				};
			}

			// Calculate the variance
			sum = 0;
			for (k = 0; k < 10; k++)
			{
				if (window_start + k + 10 * j < num_data_points)
				{
					sum += pow(data[1][window_start + k + 10 * j] - mean, 2.0);
				}
				else
				{
					sum += pow(data[1][window_start + k + 10 * j - num_data_points] - mean, 2.0);
				}
			};
			variance_sum += sqrt(sum / 9);
		}
		if ((variance_sum / num_segments) < 0.5 * F_sigma)
		{
			if (DEBUG_CANDIDATES)
			{
				printf("Too small a variance");
			}
#if MONTE_CARLO
			mc_fail_array[2]++;
#endif
			return 0;
		}
		
		/*
		 Calculate the number of points in the whole data set within 1 sigma_f_base of f_base [N_0-1sigma, counted in l], and between 1 and 3 sigma_f_base of f_base [N_1-3sigma, counted in m].
		 */
		for (i = 0, I = J = 0.0; i < num_data_points; i++)
		{
			if (data[1][i] < (median + F_sigma) && data[1][i] > (median - F_sigma))
			{
				// We've got a data point within 1 sigma of the median
				I += 1.0;
			}
			else if (data[1][i] < (median + 3 * F_sigma) && data[1][i] > (median - 3 * F_sigma))
			{
				// We've got a data point between 1 sigma and 3 sigma of the median. (if it's within 1 sigma, the previous statement caught it - so this statement wouldn't be called)
				J += 1.0;
			};
		};
		
		/*
		 The ratio of N_0-1sigma and N_1-3sigma should be 32/68 for a Gaussian distribution.
		 Calculate threshold delta = sigma_f_base * delta1 (*delta2 - can't do this bit), where delta_1 = max{(N_1-3sigma / N_0-1sigma) / (32/68):1}
		 sigma_f_base should be the variance of the mean, calculated in the main bit of the program.
		 */
		// Little dodge to remove the possibility of a divide by zero
		if (I > 0.0 && J > 0.0)
		{
			temp = (J / I) / (32.0 / 68.0);
		}
		else
		{
			temp = 0.0;
		};
		
		if (temp > 1.0)
		{
			delta = F_sigma * temp;
		}
		else
		{
			delta = F_sigma;
		};
		
//		printf("%f\n", delta);
		
		/*
		 Look for episodes with 3 consecutive points deviating by 5 * delta from f_base, in the decreasing magnitude side.
		 Look for episodes with 4 consecutive points deviating by 4 * delta from f_base, in the decreasing magnitude side. Make sure that these episodes haven't already been found in the above cut.

		 If we have ether 1 or 2 seperate episodes from the above 2 searches, then we have a candidate. Otherwise, return false.

		 l keeps track of the number of consecutive points greater than 5 sigma above, m for 5 sigma below, n for 4 sigma above, o for 4 sigma below.
		 
		 Keep an eye out for cases where the magnitude increases - if we get a match here, then dismiss the candidate.

		 */
		l = m = 0;
		j = k = 0;
		
		// These are minus 1 to avoid problems with the first data point - i==0.
		event_1_start = event_1_end = -1;
		event_2_start = event_2_end = -1;
		event_type = 0;
		for(i = 0; i < num_data_points; i++)
		{
			// Is this point more than 5 delta from the median in the wrong direction?
			if (data[1][i] > (median + 5 * delta))
			{
				j++;
				if (j > 3)
				{
					// If we're here, then we've got an event on the wrong side of the median - this isn't allowed, so return false.
					if (DEBUG_CANDIDATES)
					{
						printf("5 sigma event below the line");
					};
#if MONTE_CARLO
					mc_fail_array[3]++;
#endif
					return 0;
				}
			}
			else
			{
				j = 0;
			}
			
			// Is this point more than 4 delta from the median in the wrong direction?
			if (data[1][i] > (median + 4 * delta))
			{
				k++;
				if (k > 4)
				{
					// If we're here, then we've got an event on the wrong side of the median - this isn't allowed, so return false.
					if (DEBUG_CANDIDATES)
					{
						printf("4 sigma event below the line");
					};
#if MONTE_CARLO
					mc_fail_array[4]++;
#endif
					return 0;
				};
			}
			else
			{
				k = 0;
			}
			
			
			// Is this point more than 5 delta from the median?
			if (data[1][i] < (median - 5 * delta))
			{
				l++;
				// If this counter has got to 3, then we have an event
				if (l == 3)
				{
					if (event_1_start == -1)
					{
						// This is the first event candidate.
						// Put the event start as the starting data point number.
						event_1_start = i - 2;
						// Claim this as a 5 sigma event
						event_type = 1;
					}
					else if (event_1_end != -1 && event_1_end < i-2 && event_2_start == -1)
					{
						// The first event has ended, and is before the start of this event. This is the second event candidate.
						event_2_start = i - 2;
						// Claim this as a 5 sigma event
						event_type = 1;
					}
					else if (event_1_end != -1 && event_2_end != -1 && event_2_end < i-2)
					{
						// We've already started and finished both allowed event candidates, and this is a third event candidate. Return false.
						if (DEBUG_CANDIDATES)
						{
							printf("More than 2 candidate events found");
						}
#if MONTE_CARLO
						mc_fail_array[5]++;
#endif
						return 0;
					}
				}
			}
			else
			{
				if (event_1_start != -1 && event_1_end == -1 && event_type == 1)
				{
					// We've finished the first event, and the event type was 5 sigma. Fill in the end point, and zero the event type.
					event_1_end = i - 1;
					event_type = 0;
				}
				else if (event_2_start != -1 && event_2_end == -1 && event_type == 1)
				{
					// We've finished the second event, and the event type was 5 sigma. Fill in the end point, and zero the event type.
					event_2_end = i - 1;
					event_type = 0;
				};
				
				l = 0;
			};

			// Is this point more than 4 delta from the median?
			if (data[1][i] < (median - 4 * delta))
			{
				m++;
				// If this counter has got to 4, then we have an event
				if (m == 4)
				{
					if (event_1_start == -1)
					{
						// This is the first event candidate.
						// Put the event start as the starting data point number.
						event_1_start = i - 3;
						// Claim this as a 4 sigma event
						event_type = 2;
					}
					else if (event_1_end != -1 && event_1_end < i-2 && event_2_start == -1)
					{
						// The first event has ended, and is before the start of this event. This is the second event candidate.
						event_2_start = i - 2;
						// Claim this as a 4 sigma event
						event_type = 2;
					}
					else if (event_1_end != -1 && event_2_end != -1 && event_2_end < i-2)
					{
						// We've already started and finished both allowed event candidates, and this is a third event candidate. Return false.
						if (DEBUG_CANDIDATES)
						{
							printf("More than 2 candidate events found");
						}
#if MONTE_CARLO
						mc_fail_array[5]++;
#endif
						return 0;
					}
				};
			}
			else
			{
				if (event_1_start != -1 && event_1_end == -1 && event_type == 2)
				{
					// We've finished the first event, and the event type was 4 sigma. Fill in the end point, and zero the event type.
					event_1_end = i - 1;
					event_type = 0;
				}
				else if (event_2_start != -1 && event_2_end == -1 && event_type == 2)
				{
					// We've finished the second event, and the event type was 4 sigma. Fill in the end point, and zero the event type.
					event_2_end = i - 1;
					event_type = 0;
				};

				m = 0;
			}
			
		};
		// If we've got to here with no candidate events, return false.
		if (event_1_start == -1)
		{
			if (DEBUG_CANDIDATES)
			{
				printf("No candidate events found");
			}
#if MONTE_CARLO
			mc_fail_array[6]++;
#endif
			return 0;
		}
		else
		{
			// We've got a candidate - return true.
			return 1;
		}
	}
	else if (USE_HOMEBREW_ID)
	{
		// k will keep track of the number of points we've looked at
		k = 0;
		// l will count the number of points away from the mean by 5 sigma
		l = 0;
		// m will count the number of times l has reached more than a certain number in a row
		m = 0;
		status = 0;
		while (status == 0 && k < num_data_points)
		{
			if (data[1][k] > (median + 3 * data[2][k]))
			{
				status = 1;
			}
			// Looking for brighter events, so reduced magnitude
			else if (data[1][k] < (median - 8 * data[2][k]))
			{
				l++;
			}
			else
			{
				// We're looking for events that last more than 3 points in a row.
				if (l > 3)
				{
					m++;
				}
				l = 0;
			}
			k++;
		}
		// If we've only flagged one event greater than 5 sigma, then this could be a lightcurve.
		if (m == 1)
		{
			return 1;
		}
		else
		{
			if (DEBUG_CANDIDATES)
			{
				printf("Homebrew - no events found");
			}
			return 0;
		}
	}
	else if (USE_NUMROW_ID)
	{
		/*
		 Determine whether the lightcurve would be classified as "Transient" or "Continuous" - OGLEII only considers the former. Using Wozniak (2000) section 3.9
		 Count the number of points departing at least 3 sigma upwards, and the same downwards, and work out their ratio (n_up / n_down) - this should be outside the region 0.5 - 2.0 for a transient event.
		 */
		// j keeps track of the number of points below, k the number of points above.
		for (i = j = k = 0; i < num_data_points; i++)
		{
			if (data[1][i] > (median + 3 * data[2][i]))
			{
				// We've got a decrease in the flux (an increase in the magnitude)
				j++;
			}
			else if (data[1][i] < (median - 3 * data[2][i]))
			{
				// We've got an increase in the flux (a decrease in the magnitude)
				k++;
			};
		};
		if ((k / j) < 2.0 && (k / j) > 0.5)
		{
			// We have a continuous variable - dismiss it.
			if (DEBUG_CANDIDATES)
			{
				printf("Continuous variable");
			};
			return 0;
		};
		
		//
		// Get the index of the maximum value
		//
		min = gsl_stats_min_index(data[1], 1, num_data_points-1);
		
		//
		// Search to the left of the peak, is it decreasing for NUM_IN_ROW points? Also, is the next point greater than the baseline
		//
		for (i = 0; i < NUM_IN_ROW; i++)
		{
			if (!(data[1][min - i] < data[1][min - i - 1] && data[1][min - i - 1] < median))
			{
				return 0;
			};
		};
		
		//
		// Search to the right of the peak
		//
		for (i = 0; i < NUM_IN_ROW; i++)
		{
			if (!(data[1][min + i] < data[1][min + i + 1] && data[1][min + i + 1] < median))
			{
				return 0;
			};
		};
		
		//
		// If we're here, then a candidate event has been detected
		//
		return 1;
	}
	else if (USE_MACHO_ID)
	{
		//
		// Get the highest point in the dataset
		//
		min = gsl_stats_min_index(data[1], 1, num_data_points-1);
		
		
		//
		// Remove generic variable stars
		//
		i = j = k = 0;
		for (a = 0; a < num_data_points; a++)
		{
			if (data[1][a] > median - 4.0 * data[2][a])
			{
				i++;
			};

			if (data[1][a] > median - 2.5 * data[2][a])
			{
				j++;
			};
			
			if (data[1][a] > median - 1.5 * data[2][a])
			{
				k++;
			};
		};
		if (i < 0.66 * (double)num_data_points || j < 0.545 * (double)num_data_points || k < 0.31 * (double)num_data_points))
		{
			// Star is a generic variable
			return 0;
		};
		
		
		//
		// Remove transient variable stars
		//
		i = j = k = 0;
		for (a = 0; a < num_data_points; a++)
		{
			if (data[1][a] < median + 4.5 * data[2][a])
			{
				i++;
			};
			
			if (data[1][a] < median + 8.0 * data[2][a])
			{
				j++;
			};
			
			if (data[1][a] > median + 11.0 * data[2][a])
			{
				k++;
			};
		};
		if (!(i > 0.9 * num_data_points && j > 0.93 * num_data_points && k < 11))
		{
			// Star is a transient variable, not a microlensing event
			return 0;
		};
		

		//
		// Remove pulsating variable stars
		//
		i = j = k = 0;
		for (a = 0; a < num_data_points; a++)
		{ 
			if (data[1][a] > median + 3.5 * data[2][a])
			{
				i++;
			}
			else 
			{
				i = 0;
			};
			
			if (data[1][a] > median + 5.5 * data[2][a])
			{
				j++;
			}
			else 
			{
				j = 0;
			};
			
		};
		if (!(i < 5 && j < 4))
		{
			return 0;
		}
		
		
		//
		// Level 1 cuts complete, move onto levels 1.5 and 2
		//
		
		//
		// Fit a model of a microlensing event and a constant baseline at tpeak +/- 100days
		//
		i = j = k = 0;
		for (a = 0; a < num_data_points; a++)
		{
			//want to write a new array, which is all the data within 150 days of tpeak
			if (data[0][a] > data[0][min] - 150.0 && data[0][a] < data[0][min] + 150.0)
			{
				for (i = 0; i < 3; i++)
				{
					peak_data[i][j] = data[i][a];
					j++;
				}
			}
			
		}
		
		
		//
		// Fit the microlensing event
		//
		fit_model(peak_data, median, lsfit_params, lsfit_errors, j, &chi_global2, &chi2, &num_points_local);
		
		fit_period_model(peak_data, median, base_fit_params, base_fit_errors, j, &base_chi2);
		
		// chi_max is required to be below 3.6, chi_base < 30 
		chi_max = (1e-20 / pow(10.0,(-data[1][min] - 18.9) / 2.5)) * (chi2 + base_chi2);
		
		// Calculate mu (called A by MACHO) at t = tpeak (r = umin), want it less than 1.34
		A = (pow(lsfit_params[3], 2.0) + 2.0) / pow((pow(lsfit_params[3], 4.0) + (4.0 * pow(lsfit_params[3], 2.0))), 0.5);
		
		if (!(chi_max < 3.6 && base_chi2 < 30.0 && A > 1.34))
		{
			return 0;
		}

		//
		// Now the tests it must pass (level 1.5)
		//
		if (base_chi2 < 9.3 && chi2 > 3.0)
		{
			return 1;
		};
		
		if (base_chi2 < 15.0 && pow(10.0,(-data[1][min] - 18.9) / 2.5) > 2.5e-19)
		{
			return 1;
		}
		
		if (pow(10.0,(-data[1][min] - 18.9) / 2.5) > 5e-19)
		{
			return 1;
		}
		
		if (base_chi2 < 9.5 && chi2 < 4.0 && A > 10.0 && pow(10.0,(-data[1][min] - 18.9) / 2.5) > 3.0e-21)
		{
			return 1;
		}
		
		if (base_chi2 < 15.0 && chi2 < 4.0 && A > 20.0 && pow(10.0,(-data[1][min] - 18.9) / 2.5) > 3.0e-21)
		{
			return 1;
		}
		
		if (base_chi2 < 15.0  && A > 20.0 && pow(10.0,(-data[1][min] - 18.9) / 2.5) > 3.0e-20)
		{
			return 1;
		}
		
		if (A > 10.0 && pow(10.0,(-data[1][min] - 18.9) / 2.5) > 3.0e-19)
		{
			return 1;
		}
		
		// Level 2 cuts
		
		if (chi2 < 1.65 + (1.8e-4 * pow(10.0,(-data[1][min] - 18.9) / 2.5) * 1.0e-23) && pow(10.0,(-data[1][min] - 18.9) / 2.5) < 8.0e-20)
		{
			return 1;
		}
		
		if (base_chi2 < 3.75 + (6.0e-4 * pow(10.0,(-data[1][min] - 18.9) / 2.5) * 1.0e-23) && pow(10.0,(-data[1][min] - 18.9) / 2.5) < 1.0e-19)
		{
			return 1;
		}
		
		if (pow(10.0,(-data[1][min] - 18.9) / 2.5) > 1.0e-19)
		{
			return 1;
		}
		
		
		//
		// If we get to here without returning, then we don't have a candidate microlensing event. Return 0
		//
		return 0;
	}
	else if (USE_LSFIT_ID)
	{
		fit_model(data, median, lsfit_params, lsfit_errors, num_data_points, &chi_global2, &chi2, &num_points_local);
		
		// Check to see if the chi-squared is within bounds
		if (chi_global2 > LSF_MAX_CHI_GLOBAL || chi2 > LSF_MAX_CHI_LOCAL)
		{
//			printf("chi2 too large\n");
			return 0;
		};
		
		// Check to see if t_einstein is within bounds, and has a decent error
		if (lsfit_params[1] < LSF_T_EINSTEIN_MIN || lsfit_params[1] > LSF_T_EINSTEIN_MAX || lsfit_errors[1] > LSF_T_EINSTEIN_ERR)
		{
//			printf("t_einstein out of bounds\n");
			return 0;
		}
		
		// Check to see if u__min is within bounds, and has a decent error
		if (lsfit_params[3] < LSF_U_MIN_MIN || lsfit_params[3] > LSF_U_MIN_MAX || lsfit_errors[3] > LSF_U_MIN_ERR)
		{
//			printf("u_min out of bounds\n");
			return 0;
		}
		
		// Check to see if f_l is within bounds, and has a decent error
		if (lsfit_params[4] > LSF_F_L || lsfit_errors[4] > LSF_F_L_ERR)
		{
//			printf("f_l out of bounds\n");
			return 0;
		}

		// If we've passed all of these checks, then the lightcurve is a candidate. Return 1.
//		printf("Success!\n");
		return 1;
	};
	
	// We should never get to here, but if we do then return false.
	return 0;
};


// EOF