/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Functions to fit a theoretical microlensing model to the light curve data
 
 Started: 14/02/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h"
#include <gsl/gsl_errno.h>
// GSL non-linear least squares fit routines
#include <gsl/gsl_multifit_nlin.h>


/*
 Set up the structure of the data available (namely for GSL LSF, but may use it elsewhere)
 */
struct data_def
{
	// Number of data points in total, N
	unsigned int n;
	// Array of times
	double * time;
	// Array of the magnitudes
	double * magnitude;
	// Array of the errors
	double * sigma_mag;
};


/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the vector to which we want to output the result (pointed to by f).
 Function computes what the data would be if parameters are correct, compares it to the actual data, then outputs the difference.
 */
int fit_f(const gsl_vector *parameters, struct data_def *data, gsl_vector *f)
{
	unsigned int i;
	double r, r_2, mu, mu_b, mag;
	
	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def *)data)->n;
	double *time = ((struct data_def *)data)->time;
	double *magnitude = ((struct data_def *)data)->magnitude;
	double *sigma_mag = ((struct data_def *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//
	double t_peak = gsl_vector_get(parameters, 0);
	double t_einstein = gsl_vector_get(parameters, 1);
	double m_base = gsl_vector_get(parameters, 2);
	double u_min = gsl_vector_get(parameters, 3);
	double f_l = gsl_vector_get(parameters, 4);
	
	if (f_l > 1.0 && CALC_LSFIT_F_L)
	{
		gsl_vector_set(parameters, 4, 1.0);
		f_l = 1.0;
	}
	
	for (i = 0; i < n; i++)
	{
		//
		// Calculate the value for the magnitude that the current fitted parameters dictates at the current time
		//
		// Calculate r
		r = pow((pow(time[i] - t_peak, 2.0) / pow(t_einstein, 2.0)) + pow(u_min, 2.0), 0.5);
		r_2 = r * r;
		// Calculate mu
		mu = (r_2 + 2.0) / (r * pow(r_2 + 4.0, 0.5));
		// Calculate mu_b
		mu_b = mu * f_l - f_l + 1.0;
		// Finally, calculate the magnitude
		mag = m_base - 2.5 * log10(mu_b);
				
		//
		// Compare the calculated value for the magnitude with the actual magnitude,
		// and output the difference to vector f.
		//
		gsl_vector_set(f, i, (mag - magnitude[i]) / sigma_mag[i]);
	};
	
	return GSL_SUCCESS;
};

/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the matrix to which we want to output the result (pointed to by J).
 Function computes the differentials of the magnitude with respect to the various parameters, and puts them into the Jacobian J.
 */
int fit_df(const gsl_vector *parameters, void *data, gsl_matrix *J)
{
	unsigned int i = 0;
	double r, r_2, mu, mu_b;
	double p[NUM_PARAMETERS], dmu_b, dmu, dr;

	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def *)data)->n;
	double *time = ((struct data_def *)data)->time;
//	double *magnitude = ((struct data_def *)data)->magnitude;
	double *sigma_mag = ((struct data_def *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//
	double t_peak = gsl_vector_get(parameters, 0);
	double t_einstein = gsl_vector_get(parameters, 1);
//	double m_base = gsl_vector_get(parameters, 2);
	double u_min = gsl_vector_get(parameters, 3);
	double f_l = gsl_vector_get(parameters, 4);
	
	if (f_l > 1.0 && CALC_LSFIT_F_L)
	{
		gsl_vector_set(parameters, 4, 1.0);
		f_l = 1.0;
	}
	for (i = 0; i < n; i++)
	{
		//
		// Calculate the value for the magnitude that the current fitted parameters dictates at the current time
		//
		// Calculate r
		r = pow((pow(time[i] - t_peak, 2.0) / pow(t_einstein, 2.0)) + pow(u_min, 2.0), 0.5);
		r_2 = pow(r, 2.0);
		// Calculate mu
		mu = (r_2 + 2.0) / (r * pow(r_2 + 4.0, 0.5));
		// Calculate mu_b
		mu_b = f_l * (mu - 1.0) + 1.0;
		
		//
		// Calculate the differentials of mu_b with respect to each parameter - expressions analytically derived by Matthew.
		// First calculate some derivatives that will be used later, then calculate the actual output
		// 
		// Calculate dm / dmu_b - for use later
		// Note - M_E is the value of e, provided by GSL.
		dmu_b = - 2.5 * log10(M_E) / mu_b;
		// Calculate dm / dmu - for use later
		dmu = f_l;
		// Calculate dm/dr - for use later
		// dr = dmu * 2 * (pow(r_2 + 4, -0.5) - pow(r_2 + 2, 2) / r_2 * pow(r_2 + 4, 3/2));
		dr = 2.0 * r / pow(r_2 * r_2 + 4.0 * r_2, 0.5);
		dr -= (0.5 * (r_2 + 2.0) * (4.0 * r_2 * r + 8.0 * r)) / pow(r_2 * r_2 + 4 * r_2, 1.5);

		// Calculate dm / dt_peak
		p[0] = - dmu_b * dmu * dr * ((time[i] - t_peak) / (r * pow(t_einstein, 2.0)));
		// Calculate dm / dt_E
		p[1] = - dmu_b * dmu * dr * (pow((time[i] - t_peak), 2.0) / (r * pow(t_einstein, 3.0)));
		// Calculate dm / dm_base
		p[2] = 1.0;
		// Calculate dm / du_min
		p[3] = dmu_b * dmu * dr * u_min / r;
		// Calculate dm / df_l
		p[4] = dmu_b * (mu - 1.0);
		
		if (DEBUG_FIT_ALL)
		{
			printf("df loop %d	%f	%f	%f	%f	%f\n", i, p[0], p[1], p[2], p[3], p[4]);
		}
		
		//
		// Output the calculated parameters to the Jacobian matrix J.
		//
		gsl_matrix_set(J, i, 0, p[0] / sigma_mag[i]);
		gsl_matrix_set(J, i, 1, p[1] / sigma_mag[i]);
		gsl_matrix_set(J, i, 2, p[2] / sigma_mag[i]);
		gsl_matrix_set(J, i, 3, p[3] / sigma_mag[i]);
		gsl_matrix_set(J, i, 4, p[4] / sigma_mag[i]);
	};
	
	return GSL_SUCCESS;
};

/*
 This little function simply passes the work on to the two main functions - fit_f and fit_df above.
 It's here just in case it's required by the fitting program.
 */
int fit_fdf (const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
	fit_f(x, params, f);
	fit_df(x, params, J);
	
	return GSL_SUCCESS;
};

/*
 Fit the lightcurve parameters to the data
 Assumes that all the data fed into it is part of the light curve
 */
int fit_model (double data[3][N], const double average, double lsfit_params[5], double lsfit_errors[5], int num_data_points, double *chi_global2, double *chi2, int *num_points_local)
{
	double a_max, u_min, magnification, impact_parameter, t_einstein_left, t_einstein_right, t_einstein;
	int max_mag_point;
	double max_mag_time;
	int status, k, l, temp;
	
	// Set up the data definition structure. In order: n, time, magnitude, mag_error.
	struct data_def d = {num_data_points, data[0], data[1], data[2]};
	
	// Set up the array holding the values of the parameters for the fit
	double params_init[NUM_PARAMETERS];
	
	// Initialise the nonlinear least-squares fitting.
	const gsl_multifit_fdfsolver_type *T;
	gsl_multifit_fdfsolver *fit_solver;
	
	gsl_multifit_function_fdf fit_function;
	fit_function.f = &fit_f;
	fit_function.df = &fit_df;
	fit_function.fdf = &fit_fdf;
	fit_function.n = N;
	fit_function.p = NUM_PARAMETERS;
	fit_function.params = &d;
	
	// Set up the covar array, for use later to get the chi^2.
	gsl_matrix *covar = gsl_matrix_alloc(NUM_PARAMETERS, NUM_PARAMETERS);
	
	//
	// Get the maximum value of the magnitude - for t_peak later
	// It does this by finding the index of the max magnitude (the position in the array), then getting the time for that index.
	// Note that this uses the min function as opposed to the max - this is because the magnitude system works in reverse.
	//
	max_mag_point = gsl_stats_min_index(data[1], 1, num_data_points);
	max_mag_time = data[0][max_mag_point];
	
	//
	// Get the maximum magnification of the magnitude
	//
	a_max = calc_magnification(data[1][max_mag_point], average);

	//
	// From the max amplitude a_max, we can get u_min the minimum impact parameter
	//
	u_min = calc_u(a_max);
	
	/*
	The aim of this bit of code is to get an estimate for t_einstein. It does this by looking for a point greater than 5 sigma less magnitude than the error on the peak magnitude - this is the minimum value for a decent lightcurve signal.
	 Note to self: would it be better to use a larger number of sigma if the data set allows, as this would potentially give a better estimate of t_einstein?
	 
	 We do this analysis for both the left and right hand parts of the light curve - this way, if one is missing we still get an estimate for t_einstein - if they're both present then we get a better estimate.
	 */
	k = max_mag_point;
	t_einstein_left = 0.0;
	while (t_einstein_left == 0.0 && k > 0)
	{
		// Look at the previous data point
		k--;
		// If this is more than 5 sigma from the original, then it's the one we want - unless we've gone to the other side of the average magnitude
		if (data[1][k] > (data[1][max_mag_point] + 5 * data[2][max_mag_point]) && (data[1][k] < average))
		{
			// Calculate the magnification at this point
			magnification = calc_magnification(data[1][k], average);
			// Subsequently calculate the impact parameter
			impact_parameter = calc_u(magnification);
			// Hence get an estimate for t_einstein
			t_einstein_left = (data[0][max_mag_point] - data[0][k]) / (sqrt(pow(impact_parameter, 2.0) - pow(u_min, 2.0)));
		}
	}
	// Now do the same for the right-hand-side
	k = max_mag_point;
	t_einstein_right = 0.0;
	// NB: num_data_points is from the main program, and is the total number of points in the dataset.
	while (t_einstein_right == 0.0 && k < (num_data_points-1))
	{
		// Look at the next data point
		k++;
		// If this is more than 5 sigma from the original, then it's the one we want - unless we've gone to the other side of the average magnitude
		if (data[1][k] > (data[1][max_mag_point] + 5 * data[2][max_mag_point]) && (data[1][k] < average))
		{
			// Calculate the magnification at this point
			magnification = calc_magnification(data[1][k], average);
			// Subsequently calculate the impact parameter
			impact_parameter = calc_u(magnification);
			// Hence get an estimate for t_einstein
			t_einstein_right = (data[0][k] - data[0][max_mag_point]) / (sqrt(pow(impact_parameter, 2.0) - pow(u_min, 2.0)));
		}
	}


	// Now, what do we want to use for t_einstein? Left, right, or average? (i.e. what have we found?)
	if (t_einstein_left != 0.0 && t_einstein_right != 0.0)
	{
		// We've got both t_einstein_left and t_einstein_right - use an average of the two.
		t_einstein = (t_einstein_left + t_einstein_right) / 2.0;
	}
	else if (t_einstein_left != 0.0)
	{
		// We've got t_einstein_left, but not t_einstein_right - so just use the former
		t_einstein = t_einstein_left;
	}
	else if (t_einstein_right != 0.0)
	{
		// We've got t_einstein_right, but not t_einstein_left - so just use the former
		t_einstein = t_einstein_right;
	}
	else
	{
		// We've not found either - set t_einstein to 20 days as an estimate (as per supervisor)
		t_einstein = 20.0;
	}
	
	// 
	// Set the initial guess for parameters
	//
	// Guess for t_peak - having it as the maximum peak value should find t_peak, but spurious points will cause problems.
	params_init[0] = max_mag_time;
	// Guess for t_einstein - t_einstein_end - t_einstein_start
	params_init[1] = t_einstein;
	// Guess for m_base - the average value of the light curve
	params_init[2] = average;
	// Guess for u_min from before
	params_init[3] = u_min;
	// Guess for f_l - value of 1.0 assumes no contribution from unlensed stars
	params_init[4] = 1.0;
	// Pass these parameters to a GSL vector, which the fit routine can use.
	gsl_vector_view parameters = gsl_vector_view_array(params_init, NUM_PARAMETERS);
	//
	// Initialise the fit routine
	//
	T = gsl_multifit_fdfsolver_lmsder;
	fit_solver = gsl_multifit_fdfsolver_alloc(T, N, NUM_PARAMETERS);
	gsl_multifit_fdfsolver_set(fit_solver, &fit_function, &parameters.vector);
	
	//
	// Iterate the fitting routine until it either dies, or has gone through MAX_FIT_ITERATIONS.
	//
	k = 0;
	if (DEBUG_FIT)
	{
		printf("\n\n\nLoop	t_peak		t_einstein	m_base		u_min		f_l\n");
		printf("0	%f	%f	%f	%f	%f\n",
			   params_init[0],
			   params_init[1],
			   params_init[2],
			   params_init[3],
			   params_init[4]
			   );		
	}
	do
	{
		k++;
		status = gsl_multifit_fdfsolver_iterate(fit_solver);
		if (status && DEBUG_FIT)
		{
			printf ("Iterate error: %s\n", gsl_strerror (status));
		}
		
		status = gsl_multifit_test_delta(fit_solver->dx, fit_solver->x, 1e-4, 1e-4);
		if (status && DEBUG_FIT)
		{
			printf ("Delta test error: %s\n", gsl_strerror (status));
		}
		
		if (DEBUG_FIT)
		{
			printf("%d	%f	%f	%f	%f	%f\n",
				   k,
				   gsl_vector_get(fit_solver->x, 0),
				   gsl_vector_get(fit_solver->x, 1),
				   gsl_vector_get(fit_solver->x, 2),
				   gsl_vector_get(fit_solver->x, 3),
				   gsl_vector_get(fit_solver->x, 4)
				   );
		}
	}
	while (status == GSL_CONTINUE && k < MAX_FIT_ITERATIONS);
	
	// Now run the covar routine, to allow us to get the fit error
	gsl_multifit_covar(fit_solver->J, 0.0, covar);
	
	//
	// Output the final values from the fit
	// Note that this is fit_solver->x, NOT fit_solver->parameters.
	//
	#define FIT(i) gsl_vector_get(fit_solver->x, i)
	lsfit_params[0] = FIT(0);
	lsfit_params[1] = FIT(1);
	lsfit_params[2] = FIT(2);
	lsfit_params[3] = FIT(3);
	lsfit_params[4] = FIT(4);
	
	//
	// Output the errors on the fitting results
	//
	#define ERROR(i) sqrt(gsl_matrix_get(covar, i, i))
	lsfit_errors[0] = ERROR(0);
	lsfit_errors[1] = ERROR(1);
	lsfit_errors[2] = ERROR(2);
	lsfit_errors[3] = ERROR(3);
	lsfit_errors[4] = ERROR(4);
	
	//
	// Calculate the global chi
	//
	*chi_global2 = 0;
	for(k = 0; k < num_data_points; k++)
	{
		*chi_global2 += (gsl_vector_get(fit_solver->f, k) * gsl_vector_get(fit_solver->f, k));
	};
	*chi_global2 = *chi_global2 / (num_data_points - NUM_PARAMETERS);

	//
	// Calculate the local chi
	// Select events which are l sigma above the median near the event (start at event, work backwards, then return to start and work forwards)
	//
	l = 1;
	// Do the before bit
	k = max_mag_point;
	status = 0;
	// Don't forget to do the chi for this point!
	*chi2 = pow(gsl_vector_get(fit_solver->f, k), 2.0);
	*num_points_local = 1;
	while (status == 0 && k > 0)
	{
		// Look at the previous data point
		k--;
		// If this is more than 2 sigma from the median, then we want to count it
		if (data[1][k] < (average - l * (data[2][k])))
		{
			*chi2 += pow(gsl_vector_get(fit_solver->f, k), 2.0);
			*num_points_local = *num_points_local + 1;
		}
		else
		{
			// We've got to a point outside where we're looking, so stop the search
			status = 1;
		}
	}
	// Now do the same for the right-hand-side
	k = max_mag_point;
	status = 0;
	// NB: num_data_points is from the main program, and is the total number of points in the dataset.
	while (status == 0 && k < (num_data_points-1))
	{
		// Look at the next data point
		k++;
		// If this is more than 2 sigma from the median, then we want to count it
		if (data[1][k] < (average - l * (data[2][k])))
		{
			*chi2 += pow(gsl_vector_get(fit_solver->f, k), 2.0);
			*num_points_local = *num_points_local + 1;
		}
		else
		{
			// We've got to a point outside where we're looking, so stop the search
			status = 1;
		}
	};
	*chi2 = *chi2 / ((float)(*num_points_local - NUM_PARAMETERS));

	// Free the LSF memory
	gsl_multifit_fdfsolver_free (fit_solver);

	return 1;
}


// EOF