/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Functions to fit a straight line to the model
 
 Started: 05/05/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h"
#include <gsl/gsl_errno.h>
// GSL non-linear least squares fit routines
#include <gsl/gsl_multifit_nlin.h>


/*
 Set up the structure of the data available (namely for GSL LSF, but may use it elsewhere)
 */
struct data_def_2
{
	// Number of data points in total, N
	unsigned int n;
	// Array of times
	double * time;
	// Array of the magnitudes
	double * magnitude;
	// Array of the errors
	double * sigma_mag;
};


/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the vector to which we want to output the result (pointed to by f).
 Function computes what the data would be if parameters are correct, compares it to the actual data, then outputs the difference.
 */
int fit_base_f(const gsl_vector *parameters, void *data, gsl_vector *f)
{
	unsigned int i;
	double mag, x;
	
	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def_2 *)data)->n;
	double *time = ((struct data_def_2 *)data)->time;
	double *magnitude = ((struct data_def_2 *)data)->magnitude;
	double *sigma_mag = ((struct data_def_2 *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//

	double m_b = gsl_vector_get(parameters, 0);

	
	for (i = 0; i < n; i++)
	{

		// Calculate the magnitude
		mag = m_b;
		//
		// Compare the calculated value for the magnitude with the actual magnitude,
		// and output the difference to vector f.
		//
		gsl_vector_set(f, i, (mag - magnitude[i]) / sigma_mag[i]);
	};
	
	return GSL_SUCCESS;
};

/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the matrix to which we want to output the result (pointed to by J).
 Function computes the differentials of the magnitude with respect to the various parameters, and puts them into the Jacobian J.
 */
int fit_base_df(const gsl_vector *parameters, void *data, gsl_matrix *J)
{
	unsigned int i = 0;
	double x, y;
	double p[NUM_BASE_PARAMETERS];
	
	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def_2 *)data)->n;
	double *time = ((struct data_def_2 *)data)->time;
	//	double *magnitude = ((struct data_def_2 *)data)->magnitude;
	double *sigma_mag = ((struct data_def_2 *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//
	double m_b = gsl_vector_get(parameters, 0);

	
	for (i = 0; i < n; i++)
	{

		
		// Calculate dm / dm_b
		p[0] = 0.0;


		if (DEBUG_BASE_FIT_ALL)
		{
			printf("df loop %d	%f\n", i, p[0]);
		}
		
		//
		// Output the calculated parameters to the Jacobian matrix J.
		//
		gsl_matrix_set(J, i, 0, p[0] / sigma_mag[i]);

	};
	
	return GSL_SUCCESS;
};

/*
 This little function simply passes the work on to the two main functions - fit_f and fit_df above.
 It's here just in case it's required by the fitting program.
 */
int fit_base_fdf (const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
	fit_base_f(x, params, f);
	fit_base_df(x, params, J);
	
	return GSL_SUCCESS;
};

/*
 Fit the lightcurve parameters to the data
 Assumes that all the data fed into it is part of the light curve
 */
int fit_base_model (double data[3][N], double average, double base_fit_params[1], double base_fit_errors[1], int num_data_points, double *base_chi2)
{
	// Set up the data definition structure. In order: n, time, magnitude, mag_error.
	struct data_def_2 d = {num_data_points, data[0], data[1], data[2]};
	
	int max_mag_point, k, status;
	
	// Set up the array holding the values of the parameters for the fit
	double params_init[NUM_BASE_PARAMETERS];
	
	
	// Initialise the nonlinear least-squares fitting.
	const gsl_multifit_fdfsolver_type *T;
	gsl_multifit_fdfsolver *fit_solver;
	
	gsl_multifit_function_fdf fit_function;
	fit_function.f = &fit_base_f;
	fit_function.df = &fit_base_df;
	fit_function.fdf = &fit_base_fdf;
	fit_function.n = N;
	fit_function.p = NUM_BASE_PARAMETERS;
	fit_function.params = &d;
	
	// Set up the covar array, for use later to get the chi^2.
	gsl_matrix *covar = gsl_matrix_alloc(NUM_BASE_PARAMETERS, NUM_BASE_PARAMETERS);
	
	//
	// Get the maximum value of the magnitude - for Mc later
	// It does this by finding the index of the max magnitude (the position in the array), then getting the time for that index.
	// Note that this uses the min function as opposed to the max - this is because the magnitude system works in reverse.
	//
	max_mag_point = gsl_stats_min_index(data[1], 1, num_data_points);
	
	
	
	// 
	// Set the initial guess for parameters
	//
	// Guess for m_b
	params_init[0] = average;


	// Pass these parameters to a GSL vector, which the fit routine can use.
	gsl_vector_view parameters = gsl_vector_view_array(params_init, NUM_BASE_PARAMETERS);
	
	//
	// Initialise the fit routine
	//
	T = gsl_multifit_fdfsolver_lmsder;
	fit_solver = gsl_multifit_fdfsolver_alloc(T, N, NUM_BASE_PARAMETERS);
	gsl_multifit_fdfsolver_set(fit_solver, &fit_function, &parameters.vector);
	
	//
	// Iterate the fitting routine until it either dies, or has gone through MAX_FIT_ITERATIONS.
	//
	k = 0;
	if (DEBUG_BASE_FIT)
	{
		printf("\n\n\nLoop	base\n");
		printf("0	%f\n",
			   params_init[0]
			   );		
	}
	do
	{
		k++;
		status = gsl_multifit_fdfsolver_iterate(fit_solver);
		if (status && DEBUG_BASE_FIT)
		{
			printf ("Iterate error: %s\n", gsl_strerror (status));
		}
		
		status = gsl_multifit_test_delta(fit_solver->dx, fit_solver->x, 1e-6, 1e-6);
		if (status && DEBUG_BASE_FIT)
		{
			printf ("Delta test error: %s\n", gsl_strerror (status));
		}
		
		if (DEBUG_BASE_FIT)
		{
			printf("%d	%f\n",
				   k,
				   gsl_vector_get(fit_solver->x, 0)
				   );
		}
	}
	while (status == GSL_CONTINUE && k < MAX_FIT_ITERATIONS);
	
	// Now run the covar routine, to allow us to get the fit error
	gsl_multifit_covar(fit_solver->J, 0.0, covar);
	
	//
	// Output the final values from the fit
	// Note that this is fit_solver->x, NOT fit_solver->parameters.
	//
#define BASE_FIT(i) base_fit_params[i] = gsl_vector_get(fit_solver->x, i)
	BASE_FIT(0);

	
	//
	// Output the errors on the fitting results
	//
#define BASE_ERROR(i) base_fit_errors[i] = sqrt(gsl_matrix_get(covar, i, i))
	BASE_ERROR(0);

	
	//
	// Calculate the chi-squared
	//
	*base_chi2 = 0;
	for(k = 0; k < num_data_points; k++)
	{
		*base_chi2 = *base_chi2 + (gsl_vector_get(fit_solver->f, k) * gsl_vector_get(fit_solver->f, k));
	};
	*base_chi2 = *base_chi2 / (float)(num_data_points - NUM_BASE_PARAMETERS);
	
	// Free the LSF memory
	gsl_multifit_fdfsolver_free (fit_solver);

	return 1;
}
// That's all, folks!