/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Functions to fit a theoretical periodic model to the light curve data
 
 Started: 14/02/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h"
#include <gsl/gsl_errno.h>
// GSL non-linear least squares fit routines
#include <gsl/gsl_multifit_nlin.h>


/*
 Set up the structure of the data available (namely for GSL LSF, but may use it elsewhere)
 */
struct data_def_2
{
	// Number of data points in total, N
	unsigned int n;
	// Array of times
	double * time;
	// Array of the magnitudes
	double * magnitude;
	// Array of the errors
	double * sigma_mag;
};


/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the vector to which we want to output the result (pointed to by f).
 Function computes what the data would be if parameters are correct, compares it to the actual data, then outputs the difference.
 */
int fit_period_f(const gsl_vector *parameters, void *data, gsl_vector *f)
{
	unsigned int i;
	double mag, x;
	
	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def_2 *)data)->n;
	double *time = ((struct data_def_2 *)data)->time;
	double *magnitude = ((struct data_def_2 *)data)->magnitude;
	double *sigma_mag = ((struct data_def_2 *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//
	double Mc = gsl_vector_get(parameters, 0);
	double Ms = gsl_vector_get(parameters, 1);
	double period = gsl_vector_get(parameters, 2);
	double t_0 = gsl_vector_get(parameters, 3);
	double m_b = gsl_vector_get(parameters, 4);
	double Mc2 = gsl_vector_get(parameters, 5);
	double Ms2 = gsl_vector_get(parameters, 6);
	double Mc3 = gsl_vector_get(parameters, 7);
	double Ms3 = gsl_vector_get(parameters, 8);
	
	for (i = 0; i < n; i++)
	{
		// Calculate the frequency
		x = (2.0 * M_PI * (time[i] - t_0)) / period;
		// Calculate the magnitude
		mag = m_b + (Mc * cos(x)) + (Ms * sin(x)) + (Ms2 * sin(2.0 * x)) + (Mc2 * cos(2.0 * x)) + (Ms3 * sin(3.0 * x)) + (Mc3 * cos(3.0 * x));
		
		//
		// Compare the calculated value for the magnitude with the actual magnitude,
		// and output the difference to vector f.
		//
		gsl_vector_set(f, i, (mag - magnitude[i]) / sigma_mag[i]);
	};
	
	return GSL_SUCCESS;
};

/*
 For Non-Linear Least Squares Fitting
 Inputs are estimates of the parameters (in array x), pointers to the actual data (in params) and the matrix to which we want to output the result (pointed to by J).
 Function computes the differentials of the magnitude with respect to the various parameters, and puts them into the Jacobian J.
 */
int fit_period_df(const gsl_vector *parameters, void *data, gsl_matrix *J)
{
	unsigned int i = 0;
	double x, y;
	double p[NUM_PERIOD_PARAMETERS];
	
	//
	// Get pointers to the actual data.
	//
	unsigned int n = ((struct data_def_2 *)data)->n;
	double *time = ((struct data_def_2 *)data)->time;
	//	double *magnitude = ((struct data_def_2 *)data)->magnitude;
	double *sigma_mag = ((struct data_def_2 *)data)->sigma_mag;
	
	//
	// Get values for the parameters
	//
	double Mc = gsl_vector_get(parameters, 0);
	double Ms = gsl_vector_get(parameters, 1);
	double period = gsl_vector_get(parameters, 2);
	double t_0 = gsl_vector_get(parameters, 3);
	//	double m_b = gsl_vector_get(parameters, 4);
	double Mc2 = gsl_vector_get(parameters, 5);
	double Ms2 = gsl_vector_get(parameters, 6);
	double Mc3 = gsl_vector_get(parameters, 7);
	double Ms3 = gsl_vector_get(parameters, 8);
	
	for (i = 0; i < n; i++)
	{
		x = (2.0 * M_PI * (time[i] - t_0)) / period;
		
		y = 2.0 * M_PI / period;
		
		// Calculate dm / dMc
		p[0] = cos(x);
		// Calculate dm / dMs
		p[1] = sin(x);
		// Calculate dm / dperiod
		p[2] = (Mc * (x / period) * sin(x)) - (Ms * (x / period) * cos(x)) + (2.0 * Mc2 * (x / period) * sin(2.0 * x)) - (2.0 * Ms2 * (x / period) * cos(2.0 * x)) + (3.0 * Mc3 * (x / period) * sin(3.0 * x)) - (3.0 * Ms3 * (x / period) * cos(3.0 * x));
		// Calculate dm / dt_0
		p[3] = (Mc * y * sin(x)) - (Ms * y * cos(x)) + (2.0 * Mc2 * y * sin(2.0 * x)) - (2.0 * Ms2 * y * cos(2.0 * x)) + (3.0 * Mc3 * y * sin(3.0 * x)) - (3.0 * Ms3 * y * cos(3.0 * x));
		// Calculate dm / dm_b
		p[4] = 1.0;
		// Calculate dm / dMc2
		p[5] = cos(2.0 * x);
		// Calculate dm / dMs2
		p[6] = sin(2.0 * x);
		// Calculate dm / dMc3
		p[7] = cos(3.0 * x);
		// Calculate dm / dMs3
		p[8] = sin(3.0 * x);

		if (DEBUG_PERIOD_FIT_ALL)
		{
			printf("df loop %d	%f	%f	%f	%f	%f	%f	%f\n", i, p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8]);
		}
		
		//
		// Output the calculated parameters to the Jacobian matrix J.
		//
		gsl_matrix_set(J, i, 0, p[0] / sigma_mag[i]);
		gsl_matrix_set(J, i, 1, p[1] / sigma_mag[i]);
		gsl_matrix_set(J, i, 2, p[2] / sigma_mag[i]);
		gsl_matrix_set(J, i, 3, p[3] / sigma_mag[i]);
		gsl_matrix_set(J, i, 4, p[4] / sigma_mag[i]);
		gsl_matrix_set(J, i, 5, p[5] / sigma_mag[i]);
		gsl_matrix_set(J, i, 6, p[6] / sigma_mag[i]);
		gsl_matrix_set(J, i, 7, p[7] / sigma_mag[i]);
		gsl_matrix_set(J, i, 8, p[8] / sigma_mag[i]);
	};
	
	return GSL_SUCCESS;
};

/*
 This little function simply passes the work on to the two main functions - fit_f and fit_df above.
 It's here just in case it's required by the fitting program.
 */
int fit_period_fdf (const gsl_vector *x, void *params, gsl_vector *f, gsl_matrix *J)
{
	fit_period_f(x, params, f);
	fit_period_df(x, params, J);
	
	return GSL_SUCCESS;
};

/*
 Fit the lightcurve parameters to the data
 Assumes that all the data fed into it is part of the light curve
 */
int fit_period_model (double data[3][N], double average, double period, double period_fit_params[7], double period_fit_errors[7], int num_data_points, double *period_chi2)
{
	// Set up the data definition structure. In order: n, time, magnitude, mag_error.
	struct data_def_2 d = {num_data_points, data[0], data[1], data[2]};
	
	int max_mag_point, k, status;
	
	// Set up the array holding the values of the parameters for the fit
	double params_init[NUM_PERIOD_PARAMETERS];
	
	
	// Initialise the nonlinear least-squares fitting.
	const gsl_multifit_fdfsolver_type *T;
	gsl_multifit_fdfsolver *fit_solver;
	
	gsl_multifit_function_fdf fit_function;
	fit_function.f = &fit_period_f;
	fit_function.df = &fit_period_df;
	fit_function.fdf = &fit_period_fdf;
	fit_function.n = N;
	fit_function.p = NUM_PERIOD_PARAMETERS;
	fit_function.params = &d;
	
	// Set up the covar array, for use later to get the chi^2.
	gsl_matrix *covar = gsl_matrix_alloc(NUM_PERIOD_PARAMETERS, NUM_PERIOD_PARAMETERS);
	
	//
	// Get the maximum value of the magnitude - for Mc later
	// It does this by finding the index of the max magnitude (the position in the array), then getting the time for that index.
	// Note that this uses the min function as opposed to the max - this is because the magnitude system works in reverse.
	//
	max_mag_point = gsl_stats_min_index(data[1], 1, num_data_points);
	
	
	
	// 
	// Set the initial guess for parameters
	//
	// Guess for Mc.
	params_init[0] = (data[1][max_mag_point] - average);
	// Guess for Ms
	params_init[1] = 0.0;
	// Guess for period (from scargle)
	params_init[2] = period;
	// Guess for t_0 (the phase) - use the max mag time
	//	params_init[3] = data[0][0];
	params_init[3] = data[0][max_mag_point];
	// Guess for m_b
	params_init[4] = average;
	// Guess for Mc2.
	params_init[5] = 0.0;
	// Guess for Ms2
	params_init[6] = 0.0;
	// Guess for Mc3.
	params_init[7] = 0.0;
	// Guess for Ms3
	params_init[8] = 0.0;

	// Pass these parameters to a GSL vector, which the fit routine can use.
	gsl_vector_view parameters = gsl_vector_view_array(params_init, NUM_PERIOD_PARAMETERS);
	
	//
	// Initialise the fit routine
	//
	T = gsl_multifit_fdfsolver_lmsder;
	fit_solver = gsl_multifit_fdfsolver_alloc(T, N, NUM_PERIOD_PARAMETERS);
	gsl_multifit_fdfsolver_set(fit_solver, &fit_function, &parameters.vector);
	
	//
	// Iterate the fitting routine until it either dies, or has gone through MAX_FIT_ITERATIONS.
	//
	k = 0;
	if (DEBUG_PERIOD_FIT)
	{
		printf("\n\n\nLoop	Mc		Ms		period		phase		m_b		Mc2		Ms2		Mc3		Ms3\n");
		printf("0	%f	%f	%f	%f	%f\n",
			   params_init[0],
			   params_init[1],
			   params_init[2],
			   params_init[3],
			   params_init[4],
			   params_init[5],
			   params_init[6],
			   params_init[7],
			   params_init[8]
			   );		
	}
	do
	{
		k++;
		status = gsl_multifit_fdfsolver_iterate(fit_solver);
		if (status && DEBUG_PERIOD_FIT)
		{
			printf ("Iterate error: %s\n", gsl_strerror (status));
		}
		
		status = gsl_multifit_test_delta(fit_solver->dx, fit_solver->x, 1e-6, 1e-6);
		if (status && DEBUG_PERIOD_FIT)
		{
			printf ("Delta test error: %s\n", gsl_strerror (status));
		}
		
		if (DEBUG_PERIOD_FIT)
		{
			printf("%d	%f	%f	%f	%f	%f	%f	%f	%f	%f\n",
				   k,
				   gsl_vector_get(fit_solver->x, 0),
				   gsl_vector_get(fit_solver->x, 1),
				   gsl_vector_get(fit_solver->x, 2),
				   gsl_vector_get(fit_solver->x, 3),
				   gsl_vector_get(fit_solver->x, 4),
				   gsl_vector_get(fit_solver->x, 5),
				   gsl_vector_get(fit_solver->x, 6),
				   gsl_vector_get(fit_solver->x, 7),
				   gsl_vector_get(fit_solver->x, 8)
				   );
		}
	}
	while (status == GSL_CONTINUE && k < MAX_FIT_ITERATIONS);
	
	// Now run the covar routine, to allow us to get the fit error
	gsl_multifit_covar(fit_solver->J, 0.0, covar);
	
	//
	// Output the final values from the fit
	// Note that this is fit_solver->x, NOT fit_solver->parameters.
	//
#define PERIOD_FIT(i) period_fit_params[i] = gsl_vector_get(fit_solver->x, i)
	PERIOD_FIT(0);
	PERIOD_FIT(1);
	PERIOD_FIT(2);
	PERIOD_FIT(3);
	PERIOD_FIT(4);
	PERIOD_FIT(5);
	PERIOD_FIT(6);
	PERIOD_FIT(7);
	PERIOD_FIT(8);
	
	//
	// Output the errors on the fitting results
	//
#define PERIOD_ERROR(i) period_fit_errors[i] = sqrt(gsl_matrix_get(covar, i, i))
	PERIOD_ERROR(0);
	PERIOD_ERROR(1);
	PERIOD_ERROR(2);
	PERIOD_ERROR(3);
	PERIOD_ERROR(4);
	PERIOD_ERROR(5);
	PERIOD_ERROR(6);
	PERIOD_ERROR(7);
	PERIOD_ERROR(8);
	
	//
	// Calculate the chi-squared
	//
	*period_chi2 = 0;
	for(k = 0; k < num_data_points; k++)
	{
		*period_chi2 = *period_chi2 + (gsl_vector_get(fit_solver->f, k) * gsl_vector_get(fit_solver->f, k));
	};
	*period_chi2 = *period_chi2 / (float)(num_data_points - NUM_PERIOD_PARAMETERS);
	
	// Free the LSF memory
	gsl_multifit_fdfsolver_free (fit_solver);

	return 1;
}
// That's all, folks!