/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Various functions to augment the main program
 
 Started: 05/03/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h";

/*
 Function to go through an array (more accurately, a vector) of integers, setting all of its values to zero.
 When calling, use clear_int_array(&&array, num_points).
 */
int clear_int_array(int *array, int num_points)
{
	int i;
	for (i = 0; i < num_points; i++)
	{
		*array = 0;
		array++;
	}
	return 1;
}

/*
 Function to go through an array (more accurately, a vector) of double's, setting all of its values to 0.0.
 */
int clear_double_array(double *array, int num_points)
{
	int i;
	for (i = 0; i < num_points; i++)
	{
		*array = 0.0;
		array++;
	}
	return 1;
}

//
// Prepare the magnitude array
// Mainly for Monte Carlo, but also for others
//
int prepare_magnitude_array(int num_points, double minimum, double maximum)
{
	int k;
	double stepsize;
	
	// Work out the difference between each period
	stepsize = (maximum - minimum) / (num_points - 1);
	
	// Starting at the max frequency, and decreasing it...
	for (k = 0; k < NUM_MAG_BINS; k++)
	{
		magnitude_array[k] = minimum + k * stepsize;
		//		printf("%d	%f\n", k, magnitude_array[k]);
	};
	
	return 1;
}

//
// Find the key for a magnitude in the magnitude array
//
int find_magnitude_key(double magnitude)
{
	int k;
	
	// Starting at the minimum frequency, and going up
	for (k = 0; k < NUM_MAG_BINS; k++)
	{
		// Check to see if the magnitude has dropped below the one we're looking at in the table
		if (magnitude < magnitude_array[k])
		{
			// If it is, then return this value
			return k;
		}
	};
	
	// If we're here, then something's gone wrong. Return -1.
	return -1;
}

//
// Find the key for a time in the time array.
// If the time is not in the array already, then add it.
// Note that both the time_array and the counter num_times_found are global variables.
//
#if (COUNT_NUM_TIMES || GENERATE_ERROR_TABLE || MONTE_CARLO || SEASON_MAG_HISTOGRAM)
int find_time_key(double time)
{
	int i;
	
	//
	// Run through each of the known times, comparing them with the requested time
	//
	for (i = 0; i < num_times_found; i++)
	{
		// If the time is already accounted for, then we want to return that.
		if (time_array[i] == time)
		{
			return i;
		};
	};
	// If we're here, then it's not currently in the array, so add it to the end 
	time_array[num_times_found] = time;
	num_times_found++;

	return num_times_found - 1;
};
#endif

/*
 ran2 function from Numerical Recipes in C (Press et. al., 1992), tidied to be easily readable.
 Long period(>2√ó10^18) random number generator of L‚ÄôEcuyer with Bays-Durham shuffle and added safeguards. Returns a uniform random deviate between 0.0 and 1.0 (exclusive of the end point values). Call with idum a negative integer to initialize; thereafter, do not alter idum between successive deviates in a sequence. RNMX should approximate the largest floating value that is less than 1. 
*/
float get_random_float(long *idum)
{ 
#define IM1 2147483563 
#define IM2 2147483399 
#define AM (1.0/IM1) 
#define IMM1 (IM1-1) 
#define IA1 40014 
#define IA2 40692 
#define IQ1 53668 
#define IQ2 52774 
#define IR1 12211 
#define IR2 3791 
#define NTAB 32 
#define NDIV (1+IMM1/NTAB) 
#define EPS 1.2e-7 
#define RNMX (1.0-EPS)
	
	int j; 
	long k; 
	static long idum2 = 123456789; 
	static long iy = 0; 
	static long iv[NTAB]; 
	float temp; 

	// Do we want to initialize the random number generator?
	if (*idum <= 0)
	{
		// Initialize. 
		if (-(*idum) < 1)
		{
			// Be sure to prevent idum=0. 
			*idum = 1; 
		}
		else
		{
			*idum =- (*idum); 
		};
		idum2 = (*idum); 
		for (j = NTAB + 7; j >= 0; j--)
		{
			// Load the shuffle table (after 8 warm-ups).
			k = (*idum) / IQ1; 
			*idum = IA1 * (*idum - k * IQ1) - k * IR1; 
			if(*idum < 0)
			{
				*idum += IM1;
			};
			if(j < NTAB)
			{
				iv[j] = *idum; 
			};
		} ;
		iy=iv[0]; 
	};
	// Start here when not initializing. 
	k = (*idum) / IQ1; 
	// Compute idum = (IA1 * idum) % IM1 without overflows by Schrage‚Äôs method.
	*idum = IA1 * (*idum - k * IQ1) - k * IR1;
	if (*idum < 0)
	{
		*idum += IM1; 
	};
	k = idum2 / IQ2;
	// Compute idum2 = (IA2 * idum) % IM2 likewise.
	idum2 = IA2 * (idum2 - k * IQ2) - k * IR2; 
	if (idum2 < 0)
	{
		idum2 += IM2; 
	};
	// Will be in the range 0..NTAB-1.
	j = iy / NDIV;
	// Here idum is shuffled, idum and idum2 are combined to generate output.
	iv[j] = *idum; 
	iy = iv[j] - idum2;
	if (iy < 1)
	{
		iy += IMM1; 
	};
	if ((temp = AM * iy) > RNMX)
	{
		// Because users don‚Äôt expect endpoint values. 
		return RNMX; 
	}
	else
	{
		return temp; 
	};
};
	  

// EOF