/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 This is the master file for the project. It's rather complex. See Appendix B.4 in the report.
 
 Started: 14/02/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

/*
 Include various functions
 */
//
// Standard C headers
//
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include <string.h>
// Include the directory header file - may not work on Windows
#include <dirent.h>

//
// GSL headers
//
// GSL statistics routines
#include <gsl/gsl_statistics_double.h>
// GSL sort routines - used to get the median
#include <gsl/gsl_sort.h>
// GSL vector routines
#include <gsl/gsl_vector.h>
// GSL matrix routines
#include <gsl/gsl_matrix.h>
// GSL mathematic routines
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>

//
// GNUPlot Interface, written by N. Devillard - see http://ndevilla.free.fr/gnuplot/
//
#include "gnuplot_i.h"
#include "gnuplot_i.c"

//
// Include config
//
#include "config.h"


/*
 Setting variables here that can be accessed globally - not the best thing, but it'll have to do for now.
 */

// This is here as if it's elsewhere it gives a bus error. Need to figure out why this is at some point...
double frequency[N_FREQUENCIES];
// Array to hold the magnitude boundaries, for Monte Carlo error generating. Acts as an index for the second layer of error_array.
double magnitude_array[NUM_MAG_BINS];


#if (COUNT_NUM_TIMES || GENERATE_ERROR_TABLE || MONTE_CARLO || SEASON_MAG_HISTOGRAM)
// Vars for the number of measurement times
double time_array[NUM_OBSERVE_TIMES];
int num_times_found = 0;
int time_id;
#endif
#if (GENERATE_ERROR_TABLE || MONTE_CARLO)
// Arrays to hold stats on the errors, for monte carlo
// (time --> magnitude -->) error sum & number of errors summed
#if INTERPOLATE_ERROR_TABLE
double error_array[NUM_OBSERVE_TIMES][NUM_MAG_INTERPOLATED];
int error_counter_array[NUM_OBSERVE_TIMES][NUM_MAG_INTERPOLATED];
double magnitude_interp_array[NUM_MAG_INTERPOLATED];
#else
double error_array[NUM_OBSERVE_TIMES][NUM_MAG_BINS];
int error_counter_array[NUM_OBSERVE_TIMES][NUM_MAG_BINS];
#endif
int mag_id;
#endif

#if MONTE_CARLO
// Array for reasons for candidate failing
int mc_fail_array[10];
#endif

/*
 Include the other homemade C files
 */
#include "functions.c"
#include "lens_theory.c"

#if IDENTIFY_CANDIDATES
#include "candidates.c"
// #include "fit_base.c"
#endif

#if (CALC_LSFIT || USE_LSFIT_ID)
#include "fit.c"
#endif

#if CALC_PERIODS
#include "period.c"
#endif

#if CALC_PERIOD_FIT
#include "fit_period.c"
#endif

#if (GET_MIN_MAX_MAG || COUNT_NUM_TIMES || GENERATE_ERROR_TABLE || MONTE_CARLO || INTERPOLATE_ERROR_TABLE)
#include "montecarlo.c"
#endif

/*
 Now, on to the program itself
 */
int main (int argc, const char * argv[])
{ 
	printf("Program started\n");
	/*
	 Establish the main variables
	 */
	//
	// File handling
	//
	//	char directory[200] = INPUT_DIRECTORY;
	char directory[200];
	sprintf(directory, "%s/", argv[1]);
	//	char output_directory[200] = OUTPUT_DIRECTORY;
	char output_directory[200];
	sprintf(output_directory, "%s-graphs/", argv[1]);
	char error_output_directory[200];
	sprintf(error_output_directory, "%s-errors/", argv[1]);
	char montecarlo_output_directory[200];
	sprintf(montecarlo_output_directory, "%s-mc_lost/", argv[1]);
	char filename[400];
	DIR *directory_pntr;
	struct dirent *file;
	FILE *data_file, *output_file, *output_file2;
	
	//
	// Data archives
	//
	// 'data' holds the main set of data from the star file. data2 is for copies of the data, if need be.
	double data[3][N], data2[3][N];
	double *data_ptr = data[0];
	// Define a general array that can hold histograms of the data
	int histogram_data[N];
	// Zero the histogram_data array
	clear_int_array(histogram_data, N);
	
	//
	// Cleaning data
	//
	// array_order is used for reordering the data into time order.
	unsigned int array_order[N];
	unsigned int *array_order_ptr = array_order;
	// luminosity is for data sorting and getting the mean. It holds a copy of the data in data[1][N].
	double luminosity[N];
	double *luminosity_ptr = luminosity;
	double start_time, end_time;
	
	//
	// Statistics
	//
	double mean, median, variance;
#if GET_MIN_MAX_MAG
	// Vars for the min, max magnitudes
	double overall_min = 0.0, overall_max = 0.0;
	double min, max;
#endif
#if SEASON_MAG_HISTOGRAM
	int season[NUM_OBSERVE_TIMES + 3][NUM_MAG_BINS];
	int season_total[NUM_OBSERVE_TIMES + 3];
	int key;
	int counter;
#endif
	
#if CALC_LSFIT
	//
	// LSF to microlensing model
	//
	// Order of parameters here: t_peak, t_einstein, m_base, u_min, f_l
	double lsfit_params[5], lsfit_errors[5];
	double *lsfit_params_ptr = lsfit_params;
	double *lsfit_errors_ptr = lsfit_errors;
	double chi_global2, chi2;
	int num_points_local;
	double range_min, range_max, temp;
#endif
	double r, r_2, mu, mu_b, mag, model_time;
	
	//
	// Period finding
	//
	double period, period_chi2;
	double *frequency_ptr = frequency;
	// order of parameter: Mc and Mc2(amplitude of the cosine part), Ms and Ms2(amplitude of the sine part), period, t_0 (phase), m_b (base magnitude)
	double period_fit_params[NUM_PERIOD_PARAMETERS], period_fit_errors[NUM_PERIOD_PARAMETERS];
	int period_candidate;
	
	
	
	//constant baseline
	
	double base, base_chi2;
	
	double base_fit_params[NUM_BASE_PARAMETERS], base_fit_errors[NUM_BASE_PARAMETERS];
	
#if MONTE_CARLO
	double mc_u_min, mc_t_einstein, mc_t_peak, mc_random;
	// Seed value for the random number generator. Started off negative, as per function instructions.
	long int mc_random_seed_val = - time(0);
	long int *mc_random_seed = &mc_random_seed_val;
	int mc_mag_key_old, mc_mag_key_new;
	int mc_num_lost = 0;
	int mc_num_found = 0;
	int montecarlo[NUM_MC_POINTS][NUM_MC_POINTS][2];
	double mc_u_min_array[NUM_MC_POINTS];
	double mc_t_einstein_array[NUM_MC_POINTS];
	double difference;
	int u_min_key, t_einstein_key;
#endif
	
	//
	// Counters, and error checkers.
	//
	unsigned int h, i, k, l, m, err_check, status, candidate, subplot, continue_star;
	unsigned int l_max = 0;
	unsigned int num_data_points;
	
	//
	// Create and initiate the gnuplot session
	//
	char command[500];
	gnuplot_ctrl * plot;
	plot = gnuplot_init();
	//
	// Initialize the period finder, if need be.
	//
#if CALC_PERIODS
	// Prepare the period finder
	prepare_period_finder(frequency, N_FREQUENCIES, MIN_PERIOD, MAX_PERIOD);
#endif
#if (GENERATE_ERROR_TABLE || MONTE_CARLO || SEASON_MAG_HISTOGRAM)
	// Prepare the magnitude array
	prepare_magnitude_array(NUM_MAG_BINS, MIN_MAGNITUDE, MAX_MAGNITUDE);
#endif
#if INTERPOLATE_ERROR_TABLE
	// Prepare the magnitude array
	prepare_mag_interpolate_array(NUM_MAG_INTERPOLATED, MIN_MAGNITUDE, MAX_MAGNITUDE);
#endif
	
#if MONTE_CARLO
	// Load the error table from files
	load_error_table(error_output_directory, 1);
	
	// Create the mc_u_min_array
	difference = (MC_U_MIN_MAX - MC_U_MIN_MIN) / (NUM_MC_POINTS - 1);
	for (i = 0; i < NUM_MC_POINTS; i++)
	{
		mc_u_min_array[i] = MC_U_MIN_MIN + i * difference;
	};
	
	// Create the mc_t_einstein_array
	difference = (MC_T_EINSTEIN_MAX - MC_T_EINSTEIN_MIN) / (NUM_MC_POINTS - 1);
	for (i = 0; i < NUM_MC_POINTS; i++)
	{
		mc_t_einstein_array[i] = MC_T_EINSTEIN_MIN + i * difference;
	};
	
	//
	// Zero the montecarlo array
	//
	for (i = 0; i < NUM_MC_POINTS; i++)
	{
		for (k = 0; k < NUM_MC_POINTS; k++)
		{
			montecarlo[i][k][0] = 0;
			montecarlo[i][k][1] = 0;
		}
	}
	
	//
	// Clear the failure reason array
	//
	clear_int_array(mc_fail_array,10);
#endif
	
#if SEASON_MAG_HISTOGRAM
	for (i = 0; i < (NUM_OBSERVE_TIMES + 3); i++)
	{
		season_total[i] = 0;
		clear_int_array(season[i], NUM_MAG_BINS);
	}
#endif
	
#if PRINT_PERIODS
	// Output the period table header
	printf("Star ID		Scargle Period	LSF Period	Chi^2 / #DoF");
#endif	
	
	//
	// While loop is mainly for Monte Carlo, so extended testing looping around the data can be done.
	//
	h = 0;
	i = 0;
	while (h < NUM_LOOPS_MAX)
	{
		h++;
		
		printf("Loop number %d\n", h);
		//
		// Open the directory, and get a list of filenames
		//
		directory_pntr = opendir(directory);
		while ((file = readdir(directory_pntr)) != NULL && i < MAX_NUM_STARS)
		{
			/*
			 We have file->d_name, file->d_namlen, and possibly file->d_fileno and file->d_d_type, on a per-file basis.
			 */
			
			// Open a file, and pass the data to an array
			sprintf(filename, "%s%s", directory, file->d_name);
			data_file = fopen(filename, "r");
			if (!data_file)
			{
				printf("Failed to open %s%s", directory, file->d_name);
				return 1;
			};
			
			// Read in the file, stopping when fscanf can't get three floats from the file (err_check != 3) - this should
			// happen either at the end of the file, or when an error happens
			num_data_points = 0;
			do
			{
				err_check = fscanf(data_file, "%lf%lf%lf\n", &data[0][num_data_points], &data[1][num_data_points], &data[2][num_data_points]);
				// We seem to have some magitude measurements that are 99.999 - these are meaningless, so don't include them in the array.
				if (data[1][num_data_points] < 99.999)
				{
					num_data_points++;
				};
			} while (err_check == 3);
			// During the previous, we overcounted by 1.
			num_data_points--;
			
			// Now the data's in the array, we don't need the data file any more
			fclose(data_file);
			
			// There are some files which don't contain any data - skip them...
			if (num_data_points > 0)
			{
				/*
				 Get the minimum and maximum magnitudes. For Monte Carlo.
				 */
#if GET_MIN_MAX_MAG
				min = gsl_stats_min(data[1], 1, num_data_points-1);
				if (min < overall_min || overall_min == 0.0)
				{
					overall_min = min;
				};
				
				max = gsl_stats_max(data[1], 1, num_data_points-1);
				if (max > overall_max || overall_max == 0.0)
				{
					overall_max = max;
				};
#endif
				
#if (COUNT_NUM_TIMES || GENERATE_ERROR_TABLE)
				//
				// Run through each of the data points
				//
				for(k = 0; k < num_data_points; k++)
				{
					//
					// Counting all of the times, and adding their value to the array.
					// This is now done using the find_time_key function, which will add times to the array if they are not already there, and will return the key otherwise.
					//
					time_id = find_time_key(data[0][k]);
					
#if GENERATE_ERROR_TABLE
					// Find the key of the closest value of the magnitude to the current one
					mag_id = find_magnitude_key(data[1][k]);
					
					// Add the error of the magnitude to the error_array, and increment the corresponding counter.
					// Check to see if the function returned a decent value
					if (mag_id != -1)
					{
						error_array[time_id][mag_id] += data[2][k];
						error_counter_array[time_id][mag_id]++;
					};
#endif
				};
#endif
				
				
#if SORT_DATA
				/*
				 Sort the data into ascending time - just in case it isn't that way already.
				 We're using data2 as otherwise we'll overwrite the only copy of the data, and get into a right mess.
				 */
				// Get the permutations of the array indices and put it into array_order
				gsl_sort_index(array_order_ptr, data[0], 1, num_data_points);
				for(k = 0; k < num_data_points; k++)
				{
					data2[0][k] = data[0][array_order[k]];
					data2[1][k] = data[1][array_order[k]];
					data2[2][k] = data[2][array_order[k]];
				}
				// Pass data2 back to data
				for(k = 0; k < num_data_points; k++)
				{
					data[0][k] = data2[0][k];
					data[1][k] = data2[1][k];
					data[2][k] = data2[2][k];
				}
				/*
				 For debugging
				 printf("%d\n", num_data_points);
				 for (k = 0; k < num_data_points; k++)
				 {
					 printf("%f	%f	%f\n", data2[0][k], data[1][k], data[2][k]);
				 };
				 */
#endif
				
				/*
				 Get the minimum time. If we've sorted the data, this is the first in the array. Otherwise, it may not be the first in the file, so use the GSL function to check them all.
				 */
#if SORT_DATA
				start_time = data[0][0];
#else
				start_time = gsl_stats_min(data[0], 1, num_data_points-1);
#endif
				
#if NORMALIZE_TIMES
				/*
				 This code sets the start time to 0, and adjusts all other times accordingly.
				 */
				// Go through all of the times, subtracting start_time from it.
				for(k = 0; k < num_data_points; k++)
				{
					data[0][k] = data[0][k] - start_time;
				};
#endif
				
				//
				// Get the end time
				//
				end_time = gsl_stats_max(data[0], 1, num_data_points);
				
#if MONTE_CARLO
				/*
				 Monte Carlo insertion of events
				 Done here as this is the end of the data tidying, but before the start of the analysis of the data.
				 Note that we don't keep track of whether a MC event was added or not, as they are always added, and we don't currently care about the distribution of detections. Hence later, we're just looking at the ratio of detected events to total events.
				 */
				
				//
				// Get random values for t_peak, t_einstein, u_min
				//
				mc_random = get_random_float(mc_random_seed);
				mc_t_peak = start_time + mc_random * (end_time - start_time);
				
				mc_random = get_random_float(mc_random_seed);
				mc_t_einstein = MC_T_EINSTEIN_MIN + mc_random * (MC_T_EINSTEIN_MAX - MC_T_EINSTEIN_MIN);
				
				mc_random = get_random_float(mc_random_seed);
				mc_u_min = MC_U_MIN_MIN + mc_random * (MC_U_MIN_MAX - MC_U_MIN_MIN);
				
				//
				// Get the existing value for m_b, as this won't be changing
				//
				// First copy the data into a new vector to sort it
				for(k = 0; k < num_data_points; k++)
				{
					luminosity[k] = data[1][k];
				};
				// Then sort it
				gsl_sort(luminosity_ptr, 1, num_data_points);
				// Now we can get the median.
				median = gsl_stats_median_from_sorted_data(luminosity, 1, num_data_points);
				
				
				// Run through the data, adding in the new microlense event
				for (k = 0; k < num_data_points; k++)
				{
					//
					// Calculate the change in the magnitude needed at this point
					//
					mag = get_theory_magnitude(data[0][k], mc_t_peak, mc_t_einstein, mc_u_min, MC_F_L, data[1][k]);
					
					//
					// Alter the error on the data point appropriately
					//
					// Find the keys for the magnitudes
					mc_mag_key_old = find_magnitude_key(data[1][k]);
					mc_mag_key_new = find_magnitude_key(mag);
					// Find the key for the time
					time_id = find_time_key(data[0][k]);
					// Change the error in a scaled fasion, based off the error_table figures (ratio between new and old error)
					//				printf("Error: %f", data[2][k]);
					data[2][k] = data[2][k] * (error_array[time_id][mc_mag_key_new] / error_array[time_id][mc_mag_key_old]);
					//				printf("--> %f at magnitudes %f --> %f\n", data[2][k], data[1][k], mag);
					
					// Finally, Apply the magnitude change
					data[1][k] = mag;
				};
				
#endif
				
				
#if CALC_MEAN
				/*
				 Work out the mean luminosity for the star.
				 */
				mean = gsl_stats_mean(data[1], 1, num_data_points);
#endif
				
#if CALC_MEDIAN
				/*
				 Work out a better normal luminosity for the star - get the median.
				 This is used throughout the program.
				 */
				// First copy the data into a new vector to sort it
				for(k = 0; k < num_data_points; k++)
				{
					luminosity[k] = data[1][k];
				};
				// Then sort it
				gsl_sort(luminosity_ptr, 1, num_data_points);
				// Now we can get the median.
				median = gsl_stats_median_from_sorted_data(luminosity, 1, num_data_points);
				
				// Calculate the varience of the data from this mean.
				variance = gsl_stats_variance_m(data[1], 1.0, num_data_points, median);
#endif
				
				
				
				
				
				
#if CALC_PERIODS
				l = 0;
				m = 0;
				for(k = 0; k < num_data_points; k++)
				{
					if (data[1][k] > (median + 3 * data[2][k]))
					{
						l++;
					}
					else if (data[1][k] < (median + 3 * data[2][k]))
					{
						m++;
					}
				}
				period_candidate = 0;
				if ((m / l) < 2.0 && (m / l) > 0.5)
				{
					period_candidate = 1;
					if (PRINT_PERIODS)
					{
						// Print the star ID
						printf("\n%s", file->d_name);
					}
					
					
					period = find_period(data, num_data_points);
#if CALC_PERIOD_FIT
					// Don't do an improved fit if Scargle has returned an error
					if (period < 0)
					{
						printf("Error in finding frequency");
						period_candidate = -1;
					}
					else
					{
						//
						// Fit the model to the data
						//
						fit_period_model(data, median, period, period_fit_params, period_fit_errors, num_data_points, &period_chi2);
						// If we're outside the chisquared range, change period_candidate to -1.
						if (MAX_PERIOD_CHI == 1 && period_chi2 > MAX_PERIOD_CHI_VALUE)
						{
							period_candidate = -1;
						}
						if (PRINT_PERIODS && period_candidate == 1)
						{
							// Print the star info to screen
							printf("	%f	%f	%f", period, period_fit_params[2], period_chi2);
						}
						period = period_fit_params[2];
					}
					// Check to make sure that the period is greater than 1 day; otherwise, the program stalls.
					if (period < 1)
					{
						period_candidate = -1;
					}
#endif
					
#if CREATE_PERIOD_GRAPHS
					if (period_candidate == 1)
					{
						//
						// Create the model variable curve
						//
						sprintf(filename, "%s%s_model.dat", output_directory, file->d_name);
						output_file = fopen(filename, "w");
						if (!output_file)
						{
							printf("Failed to open %s", filename);
							return 1;
						};
						
#if (NORMALIZE_TIMES)
						model_time = 0;
#else
						model_time = start_time;
#endif
						do
						{
							// Calculate the magnitude
							mag = period_fit_params[4]
								+ period_fit_params[0] * cos(2.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
								+ period_fit_params[1] * sin(2.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
								+ period_fit_params[5] * cos(4.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
								+ period_fit_params[6] * sin(4.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
								+ period_fit_params[7] * cos(6.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
								+ period_fit_params[8] * sin(6.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2]);
							
							fprintf(output_file, "%f	%f\n", (NORMALIZE_TIMES) ? model_time + start_time : model_time, mag);
							model_time += MODEL_POINT_SPACING;
						}
						while (model_time <= end_time);
						fclose(output_file);
						
						/*
						 Plot the light curve with the theory model applied
						 Note that we just pass the filename to gnuplot rather than the array - it's easier this way (at least for now)
						 */
						gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
						gnuplot_cmd(plot, "set output \"%s%s.%s\"", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name, OUTPUT_TYPE);
						gnuplot_cmd(plot, "set size 1.0,1.0");
						gnuplot_cmd(plot, "set origin 0.0,0.0");
						gnuplot_cmd(plot, "set key off");
						gnuplot_set_xlabel(plot, "Julian Date - 2450000");
						gnuplot_set_ylabel(plot, "Magnitude");
						gnuplot_cmd(plot, "set xrange [*:*]");
						// Reverse the y-axis, to correct for the magnitude system working in reverse.
						gnuplot_cmd(plot, "set yrange [*:*] reverse");
						// Plot the data
						gnuplot_cmd(plot, "plot '%s%s' with errorbars, '%s%s_model.dat' with lines", directory, file->d_name, output_directory, file->d_name);
						gnuplot_resetplot(plot);
						
						//
						// Fold the data on this period, output the data and make a graph of the result.
						//
						sprintf(filename, "%s%s_folded.dat", output_directory, file->d_name);
						output_file = fopen(filename, "w");
						if (!output_file)
						{
							printf("Failed to open %s", filename);
							return 1;
						};
						
						// Counter l keeps track of how many periods we've gone through
						l = 0;
						for(k = 0; k < num_data_points; k++)
						{
							// If we've gone into the next period, add one to the number of periods
							// 'While' rather than 'if' as there could be a time jump of more than one period
							while ((data[0][k] - l * period) > period)
							{
								l++;
							}
							// Output the redefided time, and the luminosity
							fprintf(output_file, "%f	%f	%f\n", data[0][k] - l * period, data[1][k], data[2][k]);
						}
						
						fclose(output_file);
						
						//
						// Create the model variable curve
						//
						sprintf(filename, "%s%s_folded_model.dat", output_directory, file->d_name);
						output_file = fopen(filename, "w");
						if (!output_file)
						{
							printf("Failed to open %s", filename);
							return 1;
						};
						
						model_time = 0;
						do
						{
							// Calculate the magnitude
							mag = period_fit_params[4]
							+ period_fit_params[0] * cos(2.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
							+ period_fit_params[1] * sin(2.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
							+ period_fit_params[5] * cos(4.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
							+ period_fit_params[6] * sin(4.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
							+ period_fit_params[7] * cos(6.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2])
							+ period_fit_params[8] * sin(6.0 * M_PI * (model_time - period_fit_params[3]) / period_fit_params[2]);
							
							fprintf(output_file, "%f	%f\n", model_time, mag);
							model_time += MODEL_POINT_SPACING;
						}
						while (model_time <= period_fit_params[2]);
						fclose(output_file);
						
						
						gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
						gnuplot_cmd(plot, "set output \"%s%s_period.%s\"", output_directory, file->d_name, OUTPUT_TYPE);
						gnuplot_cmd(plot, "set size 1.0,1.0");
						gnuplot_cmd(plot, "set origin 0.0,0.0");
						gnuplot_cmd(plot, "set key off");
						gnuplot_set_xlabel(plot, "Julian Date - 2450000");
						gnuplot_set_ylabel(plot, "Magnitude");
						// Reverse the y-axis, to correct for the magnitude system working in reverse.
						gnuplot_cmd(plot, "set yrange [*:*] reverse");
						// Plot the data
						gnuplot_cmd(plot, "plot '%s%s_folded.dat' with errorbars, '%s%s_folded_model.dat' with lines", output_directory, file->d_name, output_directory, file->d_name);
						gnuplot_resetplot(plot);
						
						//
						// Plot the normalized data
						//
						sprintf(filename, "%s%s_normalized.dat", output_directory, file->d_name);
						output_file = fopen(filename, "w");
						if (!output_file)
						{
							printf("Failed to open %s", filename);
							return 1;
						};
						
						// Counter l keeps track of how many periods we've gone through
						l = 0;
						for(k = 0; k < num_data_points; k++)
						{
							
							
							// Output the redefided time, and the luminosity
							fprintf(output_file, "%f	%f	%f\n", data[0][k], data[1][k] - (period_fit_params[0] * cos(2.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2]) + period_fit_params[1] * sin(2.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2]) + period_fit_params[5] * cos(4.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2]) + period_fit_params[6] * sin(4.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2]) + period_fit_params[7] * cos(6.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2]) + period_fit_params[8] * sin(6.0 * M_PI * (data[0][k] - period_fit_params[3]) / period_fit_params[2])), data[2][k]);
						}
						
						fclose(output_file);
						
						gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
						gnuplot_cmd(plot, "set output \"%s%s_normalized.%s\"", output_directory, file->d_name, OUTPUT_TYPE);
						gnuplot_cmd(plot, "set size 1.0,1.0");
						gnuplot_cmd(plot, "set origin 0.0,0.0");
						gnuplot_cmd(plot, "set key off");
						gnuplot_set_xlabel(plot, "Julian Date - 2450000");
						gnuplot_set_ylabel(plot, "Magnitude");
						// Reverse the y-axis, to correct for the magnitude system working in reverse.
						gnuplot_cmd(plot, "set yrange [*:*] reverse");
						// Plot the data
						// Equation is: period_fit_params[0] * cos(2.0 * M_PI * x / period_fit_params[2]) + period_fit_params[1] * sin(2.0 * M_PI * x / period_fit_params[2]) + period_fit_params[4];
						gnuplot_cmd(plot, "plot '%s%s_normalized.dat' with errorbars", output_directory, file->d_name);
						gnuplot_resetplot(plot);					
					}
#endif
				};
#endif
				
				
				
				
				
				
				candidate = 0;
#if IDENTIFY_CANDIDATES
				// If this function returns 1, then we have a candidate microlensing event. Otherwise, we don't.
				candidate = identify_candidate(data, median, num_data_points);
				if (DEBUG_CANDIDATES && candidate == 0)
				{
					printf(" - %s\n", file->d_name);
				}
				
				if (candidate == 1)
				{
					//					printf("Candidate: %s\n", file->d_name);
				}
#endif
				
				
				
				
				// The odd switch here includes both the case where IDENTIFY_CANDIDATES = 0 and status = 0, but we still want CALC_LSFIT, plus where we've looked for candidates and found one, and want CALC_LSFIT.
#if CALC_LSFIT
				if (((IDENTIFY_CANDIDATES + candidate != 1) || (MONTE_CARLO && candidate == 0 && MONTE_CARLO_LOST)) && period_candidate != -1)
				{
					//
					// Fit the model
					//
					fit_model(data, median, lsfit_params, lsfit_errors, num_data_points, &chi_global2, &chi2, &num_points_local);
					
					if ((MAX_FIT_CHI == 0 || (chi_global2 < MAX_FIT_CHI_GLOBAL && chi2 < MAX_FIT_CHI_LOCAL)) || MONTE_CARLO_LOST)
					{
						
#if (PLOT_MODEL && PLOT_LSFIT_GRAPHS)
						// Get values for the theoretical curve and output them to a dat file.
						sprintf(filename, "%s%s_theory.dat", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
						output_file = fopen(filename, "w");
						if (!output_file)
						{
							printf("Failed to open %s", filename);
							return 1;
						};
						
						//
						// Calculate data for the model of the lightcurve
						//
#if (NORMALIZE_TIMES)
						model_time = 0;
#else
						model_time = start_time;
#endif
						do
						{
							// Calculate the magnitude
							mag = get_theory_magnitude(model_time, lsfit_params[0], lsfit_params[1], lsfit_params[3], lsfit_params[4], lsfit_params[2]);
							
							fprintf(output_file, "%f	%f\n", (NORMALIZE_TIMES) ? model_time + start_time : model_time - 2450000, mag);
							model_time += MODEL_POINT_SPACING;
						}
						while (model_time <= end_time);
						fclose(output_file);
						
#if NORMALIZE_TIMES
						range_min = lsfit_params[0] + start_time - 10 * lsfit_params[1];
						if (range_min < start_time)
						{
							range_min = start_time;
						}
						range_max = lsfit_params[0] + start_time + 10 * lsfit_params[1];
						if (range_max > end_time + start_time)
						{
							range_max = end_time + start_time;
						}
#else
						range_min = lsfit_params[0] - 10 * lsfit_params[1];
						if (range_min < start_time)
						{
							range_min = start_time;
						}
						range_max = lsfit_params[0] + 10 * lsfit_params[1];
						if (range_max > end_time)
						{
							range_max = end_time;
						}
#endif
						
						
						if (range_min > range_max)
						{
							temp = range_min;
							range_min = range_max;
							range_max = temp;
						}
						
						/*
						 For debugging
						 printf("Star: %s\nrange_max: %f\nrange_min: %f\n\n", file->d_name, range_max, range_min);
						 */
						
						subplot = 0;
						if (range_max > range_min)
						{
							// Get values for the theoretical curve and output them to a dat file.
							sprintf(filename, "%s%s_dif.dat", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
							output_file = fopen(filename, "w");
							if (!output_file)
							{
								printf("Failed to open %s", filename);
								return 1;
							};
							
							//
							// Calculate data for the difference plot
							//
							for (k = 0; k < num_data_points; k++)
							{
								if (data[0][k] > range_min && data[0][k] < range_max)
								{
									// We definately want a subplot if we're here
									subplot = 1;
									
									// Calculate the magnitude
									mag = get_theory_magnitude(data[0][k], lsfit_params[0], lsfit_params[1], lsfit_params[3], lsfit_params[4], lsfit_params[2]);
									
									fprintf(output_file, "%f	%f	%f\n", (NORMALIZE_TIMES) ? data[0][k] + start_time : data[0][k] - 2450000, (data[1][k] - mag), data[2][k]);
								};
							}
							fclose(output_file);
						}
#endif
					}
				};
#endif
				
				
#if MONTE_CARLO
				/*
				 Monte Carlo - has the event been detected, and flagged as a candidate?
				 Put here so that the model fitter can also be used to identify candidates
				 */
				
				//
				// Find the appropriate u_min key
				//
				u_min_key = 0;
				for (k = 0; k < NUM_MC_POINTS; k++)
				{
					// Is u_min in the array below the one we're after?
					if (mc_u_min > mc_u_min_array[k])
					{
						// If it is, then return this value
						u_min_key = k;
					}
				};
				
				//
				// Find the appropriate t_einstein key
				//
				t_einstein_key = 0;
				for (k = 0; k < NUM_MC_POINTS; k++)
				{
					// Is t_einstein in the array below the one we're after?
					if (mc_t_einstein > mc_t_einstein_array[k])
					{
						// If it is, then return this value
						t_einstein_key = k;
					}
				};
				
				//
				// Now put the event into the array.
				//
				montecarlo[u_min_key][t_einstein_key][candidate]++;
				if (candidate == 1)
				{
					mc_num_found++;
				}
				else
				{
					mc_num_lost++;
				}
#endif	
				
				
#if PLOT_LSFIT_GRAPHS
				if ((IDENTIFY_CANDIDATES + candidate != 1 || (MONTE_CARLO && candidate == 0 && MONTE_CARLO_LOST)) && period_candidate != -1)
				{
					//
					// Output the original data. Mainly for Monte Carlo.
					//
					sprintf(filename, "%s%s", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
					output_file = fopen(filename, "w");
					if (!output_file)
					{
						printf("Failed to open %s", filename);
						return 1;
					};
					for (k = 0; k < num_data_points; k++)
					{
						fprintf(output_file, "%f	%f	%f\n", data[0][k] - 2450000, data[1][k], data[2][k]);
					}
					fclose(output_file);
					
					/*
					 Plot the light curve, save it as a PNG, and reset gnuplot ready for the next loop
					 Note that we just pass the filename to gnuplot rather than the array - it's easier this way (at least for now)
					 */
					gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
					gnuplot_cmd(plot, "set output \"%s%s.%s\"", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name, OUTPUT_TYPE);
					gnuplot_cmd(plot, "set size 1.0,1.0");
					gnuplot_cmd(plot, "set origin 0.0,0.0");
					gnuplot_cmd(plot, "set key off");
					gnuplot_set_xlabel(plot, "Julian Date - 2450000");
					gnuplot_set_ylabel(plot, "Magnitude");
					gnuplot_cmd(plot, "set xrange [*:*]");
					// Reverse the y-axis, to correct for the magnitude system working in reverse.
					gnuplot_cmd(plot, "set yrange [*:*] reverse");
					// Plot the data
#if CALC_LSFIT
					// If we've calculated a fit, then put that onto the graph too
					sprintf(command, "plot '%s%s' with errorbars, '%s%s_theory.dat' with lines", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name, (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
#else
					// Otherwise, just plot the data
					sprintf(command, "plot '%s%s' with errorbars", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : directory, file->d_name);
#endif
					gnuplot_cmd(plot, command);
					gnuplot_resetplot(plot);
					
					/*
					 Do a zoomed-in plot
					 */
#if CALC_LSFIT
					gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
					gnuplot_cmd(plot, "set output \"%s%s_local.%s\"", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name, OUTPUT_TYPE);
					// Plot the data
					if (subplot)
					{
						gnuplot_cmd(plot, "set multiplot");
						gnuplot_cmd(plot, "set size 1.0,0.8");
						gnuplot_cmd(plot, "set origin 0.0,0.0");
					}
					else
					{
						gnuplot_cmd(plot, "set size 1.0,1.0");
						gnuplot_cmd(plot, "set origin 0.0,0.0");
					}
					gnuplot_cmd(plot, "set key off");
					gnuplot_set_xlabel(plot, "Julian Date - 2450000");
					gnuplot_set_ylabel(plot, "Magnitude");
					gnuplot_cmd(plot, "set xrange [%f:%f]", range_min - 2450000, range_max - 2450000);
					// Reverse the y-axis, to correct for the magnitude system working in reverse.
					gnuplot_cmd(plot, "set yrange [*:*] reverse");
					// If we've calculated a fit, then put that onto the graph too
					sprintf(command, "plot '%s%s' with errorbars, '%s%s_theory.dat' with lines", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name, (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
					gnuplot_cmd(plot, command);
					if (subplot)
					{
						gnuplot_cmd(plot, "set size 1.0,0.2");
						gnuplot_cmd(plot, "set origin 0.0,0.8");
						gnuplot_cmd(plot, "set key off");
						gnuplot_set_ylabel(plot, "Magnitude difference");
						gnuplot_cmd(plot, "set xrange [%f:%f]", range_min - 2450000, range_max - 2450000);
						// Reverse the y-axis, to correct for the magnitude system working in reverse.
						gnuplot_cmd(plot, "set yrange [*:*] reverse");
						sprintf(command, "plot '%s%s_dif.dat' with errorbars, 0 * (x) with lines", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
						gnuplot_cmd(plot, command);
						gnuplot_cmd(plot, "unset multiplot");
					}
					gnuplot_resetplot(plot);
#endif
					
					if (PLOT_MODEL && PLOT_LSFIT_GRAPHS && 0)
					{
						// Remove the temporary file we created before.
						sprintf(command, "%s%s_theory.dat", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
						status = unlink(command);
					}
					
				};
#endif
				
				
				
				
				
				
				if (CREATE_INFO_FILE && (CALC_PERIODS == 0 || period_candidate == 1) && (IDENTIFY_CANDIDATES + candidate != 1) || (MONTE_CARLO && candidate == 0 && MONTE_CARLO_LOST))
				{
					// Create an output file for the various statistics generated here, and push the various bits of data to it.
					// Potentially have this outputting a webpage with the graph from the above page embedded, and also throw together an index page at some point?
					sprintf(filename, "%s%s.txt", (MONTE_CARLO && candidate == 0) ? montecarlo_output_directory : output_directory, file->d_name);
					output_file = fopen(filename, "w");
					if (!output_file)
					{
						printf("Failed to open %s", filename);
						return 1;
					};
					fprintf(output_file, "Data File: %s\n", file->d_name);
					fprintf(output_file, "Start time (JD): %f\n", start_time);
#if CALC_MEAN
					fprintf(output_file, "Mean Luminosity: %f\n", mean);
#endif
					
#if CALC_PERIODS
					if (period != 0)
					{
						fprintf(output_file, "Period: %f\n", period);
#if CALC_PERIOD_FIT
						fprintf(output_file, "Mc1: %f\n", period_fit_params[0]);
						fprintf(output_file, "Ms1: %f\n", period_fit_params[1]);
						fprintf(output_file, "Mc2: %f\n", period_fit_params[5]);
						fprintf(output_file, "Ms2: %f\n", period_fit_params[6]);
						fprintf(output_file, "Mc3: %f\n", period_fit_params[7]);
						fprintf(output_file, "Ms3: %f\n", period_fit_params[8]);
						fprintf(output_file, "Period Chi^2 / #DoF: %f\n", period_chi2);
#endif
					};
#endif
					
#if CALC_MEDIAN
					fprintf(output_file, "Median Luminosity: %f\n", median);
#endif
					
#if CALC_LSFIT
					fprintf(output_file, "Estimate for t_peak (JD): %f ± %f\n", lsfit_params[0], lsfit_errors[0]);
					fprintf(output_file, "Estimate for t_einstein: %f ± %f\n", lsfit_params[1], lsfit_errors[1]);
					fprintf(output_file, "Estimate for m_base: %f ± %f\n", lsfit_params[2], lsfit_errors[2]);
					fprintf(output_file, "Estimate for u_min: %f ± %f\n", lsfit_params[3], lsfit_errors[3]);
					fprintf(output_file, "Estimate for f_l: %f ± %f\n", lsfit_params[4], lsfit_errors[4]);
					fprintf(output_file, "Global Chi^2 / #DoF: %f\n", chi_global2);
					fprintf(output_file, "Local Chi^2 / #DoF: %f\n", chi2);
					fprintf(output_file, "Local points considered for local chi^2: %d\n", num_points_local);
#endif
					fclose(output_file);
				};
				
				
				
				
				
				
				
#if NUM_OUTSIDE_ERROR
				/*
				 Loop through all of the values for the star luminosity, seeing if they differ from the error by a factor greater than the error
				 */
				// Clear the histogram array
				clear_int_array(histogram_data, N);
				
				l = 0;
				for(k = 0; k < num_data_points; k++)
				{
					/*
					 Switch - is the luminosity greater or lower than the mean? This determines whether we add or subtract the error.
					 There's probably a nicer way to do this, but I can't think of it offhand. Mebbe both add and subtract the error, and see if the mean is in the middle? This way will do for now, though.
					 */
					if (data[1][k] > mean)
					{
						if (1 * (data[1][k] - data[2][k]) < mean)
						{
							// This measurement is outside the error from the mean. For now, just log it.
							l++;
						};
					}
					else
					{
						if (1 * (data[1][k] + data[2][k]) > mean)
						{
							// This measurement is outside the error from the mean. For now, just log it.
							l++;
						};
					};
				};
				
				/*
				 Now we know how many points are outside the error from the mean, we can put it into the histogram_data array to create the graph later.
				 */
				histogram_data[l]++;
				
				// Keep track of the max value for l
				if (l > l_max)
				{
					l_max = l;
				};
#endif
				
#if SEASON_MAG_HISTOGRAM
				/*
				 Loop through all of the values for the star luminosity, creating magnitude histograms based on the season of the observations.
				 Currently hard-wired to three seasons, and fixed date ranges.
				 */
				for(k = 0; k < num_data_points; k++)
				{
					key = find_magnitude_key(data[1][k]);
					// Do the season tables
					if (data[0][k] < 2452300.0)
					{
						// First season
						season[0][key]++;
						season_total[0]++;
					}
					else if (data[0][k] < 2452650.0)
					{
						// Second season
						season[1][key]++;
						season_total[1]++;
					}
					else
					{
						// Third season
						season[2][key]++;
						season_total[2]++;
					}
					
					//
					// Build array of times, and do looking up of already known times to get their keys.
					//
					status = 0;
					//
					// Get the key for the time
					//
					time_id = find_time_key(data[0][k]);
					
					// Increment the appropriate entry on the time histogram.
					season[3 + time_id][key]++;
					season_total[3 + time_id]++;
				};
#endif
				
				// Keep track of how many stars we've processed
				i++;
			};
		};
		
	};
	
	printf("\n\n%d stars processed.\n\n", i);
	
	
	
	
	
	
#if SEASON_MAG_HISTOGRAM
	// Loop over the season tables
	/*
	 for (i = 0; i < (3 + NUM_OBSERVE_TIMES); i++)
	 {
		 if (i < 3)
		 {
			 sprintf(filename, "%smagnitude_histogram_season_%d.dat", error_output_directory, (i + 1));
			 printf("Season %d: %d data points\n", i, season_total[i]);
		 }
		 else
		 {
			 sprintf(filename, "%smagnitude_histogram_%f.dat", error_output_directory, time_array[i - 3]);
			 printf("Time %f: %d data points\n", time_array[i - 3], season_total[i]);
		 }
		 output_file = fopen(filename, "w");
		 if (!output_file)
		 {
			 printf("Failed to open %s", filename);
			 return 1;
		 };
		 // Create an output file, and push the data held in histogram_data to it.
		 for(k = 0; k < NUM_MAG_BINS; k++)
		 {
			 fprintf(output_file, "%f	%d\n", magnitude_array[k], season[i][k]);
		 };
		 fclose(output_file);
	 };
	 */
	
	//
	// Run over the individual time histograms (season[3+]), outputting their results
	//
	
	//
	// Time-based
	//
	// Main output file
	sprintf(filename, "%smagnitude_histogram_time.dat", error_output_directory);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	// Second output file - the medians
	sprintf(filename, "%smagnitude_histogram_time_median.dat", error_output_directory);
	output_file2 = fopen(filename, "w");
	if (!output_file2)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	for (i = 3; i < (3 + NUM_OBSERVE_TIMES); i++)
	{
		counter = 0;
		for(k = 0; k < NUM_MAG_BINS; k++)
		{
			if (season[i][k] != 0)
			{
				fprintf(output_file, "%f	%f	%f\n", time_array[i - 3], magnitude_array[k], ((float) season[i][k] / (float) season_total[i]));
				counter += season[i][k];
			}
			// Keep track of whether we're at the median value yet - update the median value to the current one if we're before the median, or just passed it within the current magnitude.
			if (counter < (season_total[i] / 2) || (counter - season[i][k]) < (season_total[i] / 2))
			{
				median = magnitude_array[k];
			};
		};
		fprintf(output_file2, "%f	%f\n", time_array[i - 3], median);
	};
	fclose(output_file2);	
	fclose(output_file);
	
	//
	// Count-based
	//
	// Main output file
	sprintf(filename, "%smagnitude_histogram_count.dat", error_output_directory);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	// Second output file - the medians
	sprintf(filename, "%smagnitude_histogram_count_median.dat", error_output_directory);
	output_file2 = fopen(filename, "w");
	if (!output_file2)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	for (i = 3; i < (3 + NUM_OBSERVE_TIMES); i++)
	{
		counter = 0;
		for(k = 0; k < NUM_MAG_BINS; k++)
		{
			if (season[i][k] != 0)
			{
				fprintf(output_file, "%d	%f	%f\n", i - 3, magnitude_array[k], ((float) season[i][k] / (float) season_total[i]));
				counter += season[i][k];
			}
			// Keep track of whether we're at the median value yet - update the median value to the current one if we're before the median, or just passed it within the current magnitude.
			if (counter < (season_total[i] / 2) || (counter - season[i][k]) < (season_total[i] / 2))
			{
				median = magnitude_array[k];
			};
		};
		fprintf(output_file2, "%d	%f\n", i - 3, median);
	};
	fclose(output_file2);	
	fclose(output_file);
#endif
	
#if GENERATE_ERROR_TABLE
	// This code outputs to individual files. Use for Monte Carlo data generation.
	for (i = 0; i < num_times_found; i++)
	{
		// Create an output file, and push the data held in histogram_data to it.
		sprintf(filename, "%serror_table_%f.dat", error_output_directory, time_array[i]);
		output_file = fopen(filename, "w");
		if (!output_file)
		{
			printf("Failed to open %s", filename);
			return 1;
		};
		for(k = 0; k < NUM_MAG_BINS; k++)
		{
			fprintf(output_file, "%f	%f	%d\n", magnitude_array[k], error_array[i][k], error_counter_array[i][k]);
		};
		fclose(output_file);
	};
	
	// This code outputs to a single file, with timestamps. Use for creating graphs.
	sprintf(filename, "%serror_table_time.dat", error_output_directory, time_array[i]);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	for (i = 0; i < num_times_found; i++)
	{
		for(k = 0; k < NUM_MAG_BINS; k++)
		{
			if (error_counter_array[i][k] != 0)
			{
				fprintf(output_file, "%f	%f	%f\n", time_array[i], magnitude_array[k], error_array[i][k] / (double)error_counter_array[i][k]);
			}
		};
	};
	fclose(output_file);
	
	// This code outputs to a single file, with unit times (i.e. one unit of time per observation). Use for creating graphs.
	sprintf(filename, "%serror_table_count.dat", error_output_directory, time_array[i]);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	for (i = 0; i < num_times_found; i++)
	{
		for(k = 0; k < NUM_MAG_BINS; k++)
		{
			if (error_counter_array[i][k] != 0)
			{
				
				fprintf(output_file, "%d	%f	%f\n", i, magnitude_array[k], error_array[i][k] / (double)error_counter_array[i][k]);
			}
		};
	};
	fclose(output_file);
#endif
	
#if MONTE_CARLO
	//
	// Print the overall efficiency
	//
	printf("Monte Carlo Results:\n	Number detected: %d\n	Number lost: %d\n	Percentage: %f\n\n", mc_num_found, mc_num_lost, (float)mc_num_found / (float)(mc_num_lost + mc_num_found));
	
	
	//
	// Create the data file for the 3D plot of detection efficiency vs. t_einstein and u_min
	//
	sprintf(filename, "%smontecarlo_3D.dat", output_directory);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	// Loop over u_min
	for (i = 0; i < NUM_MC_POINTS; i++)
	{
		// Loop over t_einstein
		for (k = 0; k < NUM_MC_POINTS; k++)
		{
			fprintf(output_file, "%f	%f	%f\n", mc_u_min_array[i], mc_t_einstein_array[k], (float)montecarlo[i][k][1] / (float)(montecarlo[i][k][0] + montecarlo[i][k][1]));
		};
	};
	fclose(output_file);
	
	//
	// Create the data file for the 2D plot of detection efficiency vs. t_einstein
	//
	sprintf(filename, "%smontecarlo_2D.dat", output_directory);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	// Loop over t_einstein
	for (i = 0; i < NUM_MC_POINTS; i++)
	{
		mc_num_found = mc_num_lost = 0;
		// Loop over u_min
		for (k = 0; k < NUM_MC_POINTS; k++)
		{
			mc_num_found += montecarlo[k][i][1];
			mc_num_lost += montecarlo[k][i][0];
		};
		if (mc_num_found + mc_num_lost != 0)
		{
			fprintf(output_file, "%f	%f\n", mc_t_einstein_array[i], (float)mc_num_found / (float)(mc_num_found + mc_num_lost));
		}
	};
	fclose(output_file);
	
	//
	// Plot the 2D graph
	//
	gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
	gnuplot_cmd(plot, "set output \"%smontecarlo_2D.%s\"", output_directory, OUTPUT_TYPE);
	gnuplot_cmd(plot, "set xlabel 't_einstein'");
	gnuplot_cmd(plot, "set ylabel 'Efficiency of detection'");
	gnuplot_cmd(plot, "set xrange [*:*]");
	gnuplot_cmd(plot, "set yrange [*:%f]", MC_U_MIN_MAX);
	gnuplot_cmd(plot, "plot '%s%s' lt 2 lw 1 pt 3 ps 1,  '%s%s' with lines", output_directory, "montecarlo_2D.dat", output_directory, "montecarlo_2D.dat");
	gnuplot_resetplot(plot);
	
	// Print reasons for failure
	if (USE_OGLE_ID)
	{
		printf("Continuous variables: %d\n", mc_fail_array[0]);
		printf("More than 2 segments in window outside range: %d\n", mc_fail_array[1]);
		printf("Too small a variance in window: %d\n", mc_fail_array[2]);
		printf("5 sigma event below the line: %d\n", mc_fail_array[3]);
		printf("4 sigma event below the line: %d\n", mc_fail_array[4]);
		printf("More than 2 candidate events found: %d\n", mc_fail_array[5]);
		printf("No candidate events found: %d\n", mc_fail_array[6]);
	}
	
#endif
	
#if GET_MIN_MAX_MAG
	printf("Minimum magnitude: %f\n", overall_min);
	printf("Maximum magnitude: %f\n", overall_max);
#endif
	
#if COUNT_NUM_TIMES
	printf("Count	Time\n");
	for (k = 0; k < num_times_found; k++)
	{
		printf("%d	%f\n", k, time_array[k]);
	};
	printf("Number of measurement times: %d\n", num_times_found);
#endif
	
#if NUM_OUTSIDE_ERROR
	// Create an output file, and push the data held in histogram_data to it.
	sprintf(filename, "%snum_outside_error.dat", output_directory);
	output_file = fopen(filename, "w");
	if (!output_file)
	{
		printf("Failed to open %s", filename);
		return 1;
	};
	for(i = 0; i <= l_max; i++)
	{
		fprintf(output_file, "%d	%d\n", i, histogram_data[i]);
	};
	fclose(output_file);
	
	// Plot the graph for the number of points outside the error from the mean.
	gnuplot_cmd(plot, "set terminal %s size %d,%d", OUTPUT_TYPE, OUTPUT_SIZE_X, OUTPUT_SIZE_Y);
	gnuplot_cmd(plot, "set output \"%snum_outside_error.%s\"", output_directory, OUTPUT_TYPE);
	gnuplot_cmd(plot, "set xlabel 'Number of points outside the error'");
	gnuplot_cmd(plot, "set ylabel 'Number of stars'");
	gnuplot_cmd(plot, "set xrange [*:*]");
	gnuplot_cmd(plot, "set yrange [*:*]");
	gnuplot_cmd(plot, "plot '%s%s' smooth frequency with histeps", output_directory, "num_outside_error.dat");
	gnuplot_resetplot(plot);
#endif
	
	
	
	
	
	
	// Close the gnuplot connection
	gnuplot_close(plot);
	
	// Close the directory
	closedir(directory_pntr);
	
	printf("\nProgram finished\n");
	
    return 0;
}


// That's all, folks!