/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Monte Carlo functions/routines
 
 Started: 21/04/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h"

#if INTERPOLATE_ERROR_TABLE
#include <gsl/gsl_spline.h>
#include <gsl/gsl_interp.h>
#endif



//
// Prepare interpolated magnitude array, if needed
//
#if INTERPOLATE_ERROR_TABLE
int prepare_mag_interpolate_array(int num_points, double minimum, double maximum)
{
	int k;
	double stepsize;
	
	// Work out the difference between each period
	stepsize = (maximum - minimum) / (num_points - 1);
	
	// Starting at the max frequency, and decreasing it...
	for (k = 0; k < NUM_MAG_BINS; k++)
	{
		magnitude_interp_array[k] = minimum + k * stepsize;
	};
	
	return 1;
}
#endif

//
// This would be the interpolater function. It's on hold for now.
//
#if INTERPOLATE_ERROR_TABLE
void interpolate(double x[], double y[], int n_in, double xout[], double yout[], int n_out)
{
    int i;
	
    // gsl_interp_accel and gsl_spline are types that have been defined in gsl/interp.h
	// Allocate space for the interpolation
    gsl_interp_accel *acc = gsl_interp_accel_alloc();
	// Set up the interpolation routine
    gsl_interp *spline = gsl_spline_alloc(gsl_interp_cspline, n_in);
    gsl_spline_init(spline, x, y, n_in);
	
    // this loops through 10000 x values between the start and end
    // of the input data 
    for (i=0; i<nout; i++) {
        xout[i] = x[0] + (double) i * (x[nin-1]-x[0])/(double) n_out;
		// the next line does the interpolation
        yout[i] = gsl_spline_eval (spline, x, y, xout[i], acc);
	}
	
	// Free up memory used by the routine
	gsl_interp_free (spline);
	gsl_interp_accel_free (acc);
};
#endif


//
// Function to take the data out of the stored error arrays, put it back into accessible arrays, and optionally normalize and interpolate it.
//
int load_error_table(char directory[200], int normalize)
{
	DIR *directory_pntr;
	struct dirent *file;
	char filename[400];
	FILE *data_file;

	int i, k, err_check, time_counter;
	double mag;
	
	// Open directory listing
	directory_pntr = opendir(directory);
	time_counter = 0;
	while ((file = readdir(directory_pntr)) != NULL && time_counter < NUM_OBSERVE_TIMES)
	{
		/*
		 We have file->d_name, file->d_namlen, and possibly file->d_fileno and file->d_d_type, on a per-file basis.
		 */
		
		// Get the file time, and put it into the time_array.
		err_check = sscanf(file->d_name, "error_table_%lf.dat", &time_array[time_counter]);
//		printf("\n\nTime: %f\n", time_array[time_counter]);
		// Only continue if we've successfully read in the time.
		if (err_check == 1)
		{
			// Open the file, and pass the data to an array
			sprintf(filename, "%s%s", directory, file->d_name);
			data_file = fopen(filename, "r");
			if (!data_file)
			{
				printf("Failed to open %s%s", directory, file->d_name);
				return 0;
			};
			
			// Read in the file, stopping when fscanf can't get three floats from the file (err_check != 3) - this should
			// happen either at the end of the file, or when an error happens
			i = 0;
			do
			{
				err_check = fscanf(data_file, "%lf%lf%d\n", &mag, &error_array[time_counter][i], &error_counter_array[time_counter][i]);
				// Check for data integrity
				if ((magnitude_array[i] > mag + 0.000001 || magnitude_array[i] < mag - 0.000001) && err_check == 3)
				{
					printf("Error: apparent data corruption in %s - have you been messing with it, or the settings for the min and max magnitudes?\nMagnitude found: %f, magnitude looked for: %f", file->d_name, mag, magnitude_array[i]);
					return 0;
				};
				i++;
			} while (err_check == 3);
			// i has overcounted by 1 - reduce it.
			i--;
			
			// Now the data's in the array, we don't need the data file any more
			fclose(data_file);
			
			// Do we need to normalize the array?
			if (normalize = 1)
			{
				for(k = 0; k < i; k++)
				{
					if (error_counter_array[time_counter][k] != 0)
					{
						error_array[time_counter][k] = error_array[time_counter][k] / error_counter_array[time_counter][k];
						error_counter_array[time_counter][k] = 1;
					}
				};
			};

			//
			// We need to check for places where the entry is zero, and 'extrapolate' to the edges of the array. In this case, extrapolate means just carry the closest value through. This could be improved, with time.
			//
			// Start at the middle, and work to the start
			for (k = (i / 2); k != 0; k--)
			{
				if (error_counter_array[time_counter][k] == 0)
				{
					error_array[time_counter][k] = error_array[time_counter][k + 1];
					error_counter_array[time_counter][k] == 1;
				};
			};
			// Start at the middle, and work to the end
			for (k = (i / 2); k < i; k++)
			{
				if (error_counter_array[time_counter][k] == 0)
				{
					error_array[time_counter][k] = error_array[time_counter][k - 1];
					error_counter_array[time_counter][k] == 1;
				};
			};
			
			
			/*
			 For debugging
			for(k = 0; k < i; k++)
			{
				printf("%f	%f	%d\n", magnitude_array[k], error_array[time_counter][k], error_counter_array[time_counter][k]);
			}
			*/
			// increment the time_counter for the next loop.
			time_counter++;
		};
	};
	
	// Return the number of times read in if suceeded.
	return (time_counter - 1);
}