/*
 PC 4182 MPhys Project - Classification of Gravitational Microlensing Light Curves
 
 Functions to find a period in the data
 
 Continuation of Matthew's program, started 16/02/06.
 
 Started: 09/03/06
 Last Modified: 12/05/06
 
 Mike Peel (http://www.mikepeel.net/) & Matthew Dennison (matthew.dennison@manchester.ac.uk)
 */

#include "config.h"
#include <gsl/gsl_errno.h>
// GSL mathematic routines
#include <gsl/gsl_math.h>
#include <math.h>


//
// Creates an array of frequencies/periods to be tested.
//
int prepare_period_finder(double *array, int num_points, double min_period, double max_period)
{
	int k;
	double period_difference;
	
	// Work out the difference between each period
	period_difference = (max_period - min_period) / (num_points - 1);
	
	// Starting at the max frequency, and decreasing it...
	for (k = 0; k < num_points; k++)
	{
		*array = 1 / (min_period + k * period_difference);
		array++;
	};
	
	return 1;
}

//
// Period finder. Returns the period.
//
double find_period(double data[3][N], int num_data_points)
{
	// Create the necessary variables
	double mean, variance, period, frequency_difference;
	double tau, X, Y;
	double A1, B1, C1, D1;
	double spectrum[N_FREQUENCIES];

	// Counters
	int i, k, l;
	int best_fit_id = 0;
	int num_frequencies = N_FREQUENCIES;

	// Create a pointer to point to the frequency arrays, and point it to the main one to start with
	double *period_pointer = &frequency[0];

	// Second frequency array
	double frequency_array[N_BATCH_FREQUENCIES];
	
	// Calculate the mean and the variance
	mean = gsl_stats_mean(data[1], 1.0, num_data_points);
	variance = gsl_stats_variance_m(data[1], 1.0, num_data_points, mean);

	// Clear the spectrum array
	clear_double_array(spectrum, N_FREQUENCIES);
	
	// Set the frequency difference to that for the standard frequency array
	frequency_difference = 1 / ((MAX_PERIOD - MIN_PERIOD) / (N_FREQUENCIES - 1));

	
	//
	// Loop over all of the frequencies, trying to fit them to the data, getting a power spectrum out
	//
	
	// Set a counter for the number of times we've gone through a frequency array
	i = 0;
	// We want to stop this loop when the frequency difference drops below the predefined value
	while ((1 / frequency_difference) > PERIOD_ACCURACY)
	{
		// If we're in the i == 1 loop, we want to change the pointer to the second array, then use the original array's values for frequencies to build the next set of frequencies
		if (i == 1)
		{
			period_pointer = &frequency_array[0];
			num_frequencies = N_BATCH_FREQUENCIES;

			// Nasty way to get round the potential problem where the best_fit_id is on the edge of the array. This needs improving so it does something logical...
			if (best_fit_id == 0)
			{
				best_fit_id++;
			};
			if (best_fit_id == (N_FREQUENCIES - 1))
			{
				best_fit_id--;
			};

			prepare_period_finder(frequency_array, N_BATCH_FREQUENCIES, 1 / frequency[best_fit_id - 1], 1 / frequency[best_fit_id + 1]);
			frequency_difference = 1 / (((1 / frequency[best_fit_id + 1]) - (1 / frequency[best_fit_id - 1])) / (N_FREQUENCIES - 1));
		};
		// If we're in a custom frequency loop, then populate the array
		if (i > 1)
		{
			// Resetting the pointer (not sure if this is needed, or if there's another way around it)
			period_pointer = &frequency_array[0];

			// Nasty way to get round the potential problem where the best_fit_id is on the edge of the array. This needs improving so it does something logical...
			if (best_fit_id == 0)
			{
				best_fit_id++;
			};
			if (best_fit_id == (N_BATCH_FREQUENCIES - 1))
			{
				best_fit_id--;
			};
			prepare_period_finder(frequency_array, N_BATCH_FREQUENCIES, 1 / frequency_array[best_fit_id - 1], 1 / frequency_array[best_fit_id + 1]);
			frequency_difference = 1 / (((1 / frequency_array[best_fit_id + 1]) - (1 / frequency_array[best_fit_id - 1])) / (N_BATCH_FREQUENCIES - 1));
		};
		// Reset the best_fit_id, and go through the array working out the best value
		best_fit_id = 0;
		for (k = 0; k < num_frequencies; k++)
		{
			//
			// Loop over all of the data points, adding a contribution from each time to X and Y
			// This uses a different frequency each time the loop above runs
			//
			X = Y = 0.0;
			for (l = 0; l < num_data_points; l++)
			{
				X += sin(4.0 * M_PI * (*period_pointer) * data[0][l]);
				Y += cos(4.0 * M_PI * (*period_pointer) * data[0][l]);
			}
			// 
			// Work out tau - the characteristic time
			//
			tau = atan(X / Y) / (4.0 * M_PI * (*period_pointer));
			
			//
			// To find N, we need to calculate a power spectrum.
			// Suddenly this is using 2 pi, not 4 pi - why?
			//
			A1 = C1 = 0.0;
			B1 = D1 = 0.0;
			for (l = 0; l < num_data_points; l++)
			{
				//
				// These two are the cos and sin terms weighted by the change in magnitude
				// at time t - tau (tau being a characteristic time?)
				//
				A1 += (data[1][l] - mean) * cos(2.0 * M_PI * (*period_pointer) * (data[0][l] - tau));
				C1 += (data[1][l] - mean) * sin(2.0 * M_PI * (*period_pointer) * (data[0][l] - tau));
				
				//
				// These two are the non-weighted cos and sin terms, squared
				//
				B1 += pow(cos(2.0 * M_PI * (*period_pointer) * (data[0][l] - tau)), 2.0);
				D1 += pow(sin(2.0 * M_PI * (*period_pointer) * (data[0][l] - tau)), 2.0);
			};
			
			//
			// This compares A1 & B1, and C1 & D1, from above.
			//
			spectrum[k] = (pow((2.0 * variance), -1.0)) * ((pow(A1, 2.0) / B1) + (pow(C1, 2.0) / D1));
			
			//
			// If this value of N is the largest, then save it as the maximum - i.e. the best fit
			// d keeps track of what frequency had the best fit
			// Changed such that it only keeps a record of which entry was the largest, rather than copying the value into a new variable - this's a bit more elegant.
			//
			if (spectrum[k] > spectrum[best_fit_id])
			{
				best_fit_id = k;
			//	printf("%.10Lf\n", 1 / *period_pointer);
			};
			
			// Move the period pointer to the next value in the array
			period_pointer++;
			//		printf("%d	%f\n", k, spectrum[k]);
			
		};
		
		i++;
	};
	
	//	printf("%d", best_fit_id);
	
	
	//
	// Find the period, 1 /f, for the frequency of best fit.
	//
	// If the value was found to the required accuracy within the first array, return that
	if (i == 1)
	{
		period = 1.0 / frequency[best_fit_id];
	}
	else
	{
		period = 1.0 / frequency_array[best_fit_id];
	}
	
	return period;
}


// EOF